# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2024, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Preview a VisualizationSet from any component with a vis_set output.
_
The VisualizationSet is often a much more detailed view of the geometry that
the component typically generates and includes features like recommended line
weights/types, display modes (eg. wireframe vs. shaded), transparency, and more.
-

    Args:
        _vis_set: VisualizationSet arguments from any Ladybug Tools component with a vis_set
            output. This can also be the path to a .vsf file that exists on this
            machine (these files are often written with the "LB Dump VisualizationSet"
            component). Lastly, this input can be a custom VisualizationSet that
            has been created with the Ladybug Tools SDK.
        legend_par_: Optional legend parameters from the "LB Legend Parameters" component,
            which will overwrite the existing legend parameters on the input
            Visualization Set.
        leg_par2d_: Optional 2D LegendParameters from the "LB Legend Parameters 2D"
            component, which will be used to customize a legend in the plane
            of the screen so that it functions like a head-up display (HUD).
            If unspecified, the VisualizationSet will be rendered with 3D
            legends in the Rhino scene much like the other native Ladybug
            Tools components.
        data_set_: Optional text or an integer to select a specific data set from analysis
            geometries within the Visualization Set. Note that this input only has
            meaning for Visualization Sets that contain multiple data sets assigned
            to the same geometry. When using an integer, this will refer to the
            index of the data set to be visualized (starting with 0). When using
            text, this will refer to the name of the data type for the data set
            to be displayed.
        viewport_: Text for the name of the Rhino viewport to which the 2D screen-oriented
            legend will be rendered. If unspecified, the 2D legend will be rendered
            in all viewports. Acceptable inputs include:
                -
                Perspective
                Top
                Bottom
                Left
                Right
                Front
                Back
                any view name that has been saved within the Rhino file

    Returns:
        vs: A VisualizationSet object that can be baked into the Rhino document by
            running "Bake" on this component or written to a standalone file
            using the "LB Dump VisualizationSet" component.
"""

from ghpythonlib.componentbase import executingcomponent as component
import Grasshopper, GhPython
import System
import Rhino
import rhinoscriptsyntax as rs

class MyComponent(component):
    
    def __init__(self):
        super(MyComponent,self).__init__()
        self.vis_con = None
        self.vs_goo = None
        self.viewport = None
    
    def RunScript(self, _vis_set, legend_par_, leg_par2d_, data_set_, viewport_):
        ghenv.Component.Name = 'LB Preview VisualizationSet'
        ghenv.Component.NickName = 'VisSet'
        ghenv.Component.Message = '1.8.2'
        ghenv.Component.Category = 'Ladybug'
        ghenv.Component.SubCategory = '4 :: Extra'
        ghenv.Component.AdditionalHelpFromDocStrings = '1'
        
        try:
            from ladybug_display.visualization import AnalysisGeometry
        except ImportError as e:
            raise ImportError('\nFailed to import ladybug_display:\n\t{}'.format(e))
        
        try:
            from ladybug_rhino.grasshopper import all_required_inputs, objectify_output
            from ladybug_rhino.preview import VisualizationSetConverter
            from ladybug_rhino.visset import VisSetGoo, process_vis_set
        except ImportError as e:
            raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))
        
        
        if all_required_inputs(ghenv.Component):
            # process the input visualization set
            if len(_vis_set) == 1:
                vis_set = _vis_set[0]
            else:
                if hasattr(_vis_set[0], 'data'):
                    arr_type = (list, tuple)
                    if isinstance(_vis_set[0].data, arr_type) and \
                            isinstance(_vis_set[0].data[0], arr_type):
                        vis_set = objectify_output(
                            'Multiple Vis Set Args', [obj.data[0] for obj in _vis_set])
                    else:
                        vis_set = objectify_output(
                            'Multiple Vis Set Args', [obj.data for obj in _vis_set])
                else:
                    vis_set = objectify_output(
                        'Multiple Vis Sets', [[obj] for obj in _vis_set])
            vis_set_obj = process_vis_set(vis_set)
            
            # process the connected legend parameters
            if legend_par_ is not None:
                for geo in vis_set_obj:
                    if isinstance(geo, AnalysisGeometry):
                        for data in geo.data_sets:
                            # override legend properties if they are specified
                            if legend_par_.min is not None:
                                data.legend_parameters.min = legend_par_.min
                            if legend_par_.max is not None:
                                data.legend_parameters.max = legend_par_.max
                            if not legend_par_.are_colors_default:
                                data.legend_parameters.colors = legend_par_.colors
                            if not legend_par_.is_segment_count_default:
                                data.legend_parameters.segment_count = \
                                    legend_par_.segment_count
                            if data.data_type is not None:
                                unit = data.unit if data.unit else data.data_type.units[0]
                                data.legend_parameters.title = \
                                    '{} ({})'.format(data.data_type.name, unit) \
                                    if not legend_par_.vertical else unit
                            # set the other genreic legend parameter properties
                            data.legend_parameters.continuous_legend = \
                                legend_par_.continuous_legend
                            data.legend_parameters.decimal_count = \
                                legend_par_.decimal_count
                            data.legend_parameters.include_larger_smaller = \
                                legend_par_.include_larger_smaller
                            data.legend_parameters.font = legend_par_.font
                            data.legend_parameters.vertical = legend_par_.vertical
                            # override any 3D geometry properties
                            if not legend_par_.is_base_plane_default:
                                data.legend_parameters.base_plane = \
                                    legend_par_.base_plane
                            if not legend_par_.is_segment_height_default:
                                data.legend_parameters.segment_height = \
                                    legend_par_.segment_height
                            if not legend_par_.is_segment_width_default:
                                data.legend_parameters.segment_width = \
                                    legend_par_.segment_width
                            if not legend_par_.is_text_height_default:
                                data.legend_parameters.text_height = \
                                    legend_par_.text_height
            # process connected 2D legend parameters
            if leg_par2d_ is None:
                leg3d, leg2d = True, False
            else:
                leg3d, leg2d = False, True
                for geo in vis_set_obj:
                    if isinstance(geo, AnalysisGeometry):
                        for data in geo.data_sets:
                            data.legend_parameters.properties_2d = leg_par2d_
            # process the active data_set_ if it is specified
            if data_set_ is not None:
                try:  # data set is an integer referring to an index
                    data_set_ = int(data_set_)
                except ValueError:
                    pass  # data set is text referring to the data type
                for geo in vis_set_obj:
                    if isinstance(geo, AnalysisGeometry):
                        if isinstance(data_set_, int):
                            if data_set_ < len(geo.data_sets):
                                geo.active_data = data_set_
                        else:
                            for i, dat in enumerate(geo.data_sets):
                                if str(dat.data_type) == data_set_:
                                    geo.active_data = i
            self.vis_con = VisualizationSetConverter(vis_set_obj, leg3d, leg2d)
            self.vs_goo = VisSetGoo(vis_set_obj)
            self.viewport = viewport_
        else:
            self.vis_con = None
            self.vs_goo = None
            self.viewport = None
        
        # return the bake-able version of the visualization set 
        return self.vs_goo
    
    def DrawViewportWires(self, args):
        try:
            if self.vis_con is not None:
                # get the DisplayPipeline from the event arguments
                display = args.Display
                
                # for each object to be rendered, pass the drawing arguments
                for draw_args in self.vis_con.draw_3d_text:
                    display.Draw3dText(*draw_args)
                for draw_args in self.vis_con.draw_mesh_wires:
                    display.DrawMeshWires(*draw_args)
                for draw_args in self.vis_con.draw_mesh_vertices:
                    display.DrawMeshVertices(*draw_args)
                for draw_args in self.vis_con.draw_point:
                    display.DrawPoint(*draw_args)
                for draw_args in self.vis_con.draw_arrow:
                    display.DrawArrow(*draw_args)
                for draw_args in self.vis_con.draw_brep_wires:
                    display.DrawBrepWires(*draw_args)
                for draw_args in self.vis_con.draw_line:
                    display.DrawLine(*draw_args)
                for draw_args in self.vis_con.draw_patterned_line:
                    display.DrawPatternedLine(*draw_args)
                for draw_args in self.vis_con.draw_patterned_polyline:
                    display.DrawPatternedPolyline(*draw_args)
                for draw_args in self.vis_con.draw_curve:
                    display.DrawCurve(*draw_args)
                for draw_args in self.vis_con.draw_circle:
                    display.DrawCircle(*draw_args)
                for draw_args in self.vis_con.draw_arc:
                    display.DrawArc(*draw_args)
                for draw_args in self.vis_con.draw_sphere:
                    display.DrawSphere(*draw_args)
                for draw_args in self.vis_con.draw_cone:
                    display.DrawCone(*draw_args)
                for draw_args in self.vis_con.draw_cylinder:
                    display.DrawCylinder(*draw_args)
                if self.viewport is None or \
                        self.viewport.lower() == args.Viewport.Name.lower():
                    for draw_args in self.vis_con.draw_2d_text:
                        display.Draw2dText(*draw_args)
                    for draw_args in self.vis_con.draw_sprite:
                        display.DrawSprite(*draw_args)
        except Exception, e:
            System.Windows.Forms.MessageBox.Show(str(e), "script error")
    
    def DrawViewportMeshes(self, args):
        try:
            if self.vis_con is not None:
                # get the DisplayPipeline from the event arguments
                display = args.Display
                
                # for each object to be rendered, pass the drawing arguments
                for draw_args in self.vis_con.draw_mesh_false_colors:
                    display.DrawMeshFalseColors(draw_args)
                for draw_args in self.vis_con.draw_mesh_shaded:
                    display.DrawMeshFalseColors(draw_args[0])
                for draw_args in self.vis_con.draw_brep_shaded:
                    display.DrawBrepShaded(*draw_args)
        except Exception, e:
            System.Windows.Forms.MessageBox.Show(str(e), "script error")
    
    def get_ClippingBox(self):
        if self.vis_con is not None:
            return self.vis_con.bbox
        else:
            return Rhino.Geometry.BoundingBox()



# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2024, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Apply a conditional statement to a data collection.
-

    Args:
        _data: A list of aligned Data Collections to be evaluated against
            the _statement.
        _statement: A conditional statement as a string (e.g. a > 25).
            _
            The variable of the first data collection should always be named 'a'
            (without quotations), the variable of the second list should be
            named 'b', and so on.
            _
            For example, if three data collections are connected to _data
            and the following statement is applied:
            '18 < a < 26 and b < 80 and c > 2'
            The resulting collections will only include values where the first
            data collection is between 18 and 26, the second collection is less
            than 80 and the third collection is greater than 2.

    Returns:
        data: A list of Data Collections that have been filtered by the _statement.
"""

ghenv.Component.Name = 'LB Apply Conditional Statement'
ghenv.Component.NickName = 'Statement'
ghenv.Component.Message = '1.8.2'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '1 :: Analyze Data'
ghenv.Component.AdditionalHelpFromDocStrings = '3'

try:
    from ladybug.datacollection import BaseCollection
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    for dat in _data:
        assert isinstance(dat, BaseCollection), '_data must be a data' \
            ' collection. Got {}.'.format(type(dat))

    data = BaseCollection.filter_collections_by_statement(
        _data, _statement)


# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2024, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Import climate data from a standard .epw file.

-

    Args:
        _epw_file: An .epw file path on your system as a string.
        
    Returns:
        location: A Ladybug Location object describing the location data in the
            weather file.
        dry_bulb_temperature: The houlry dry bulb temperature, in C.
            Note that this is a full numeric field (i.e. 23.6) and not an integer
            representation with tenths. Valid values range from 70 C to
            70 C. Missing value for this field is 99.9.
        dew_point_temperature: The hourly dew point temperature, in C.
            Note that this is a full numeric field (i.e. 23.6) and not an integer
            representation with tenths. Valid values range from 70 C to
            70 C. Missing value for this field is 99.9.
        relative_humidity: The hourly Relative Humidity in percent.
            Valid values range from 0% to 110%. Missing value for this field is 999.
        wind_speed: The hourly wind speed in m/sec.
            Values can range from 0 to 40. Missing value is 999.
        wind_direction: The hourly wind direction in degrees.
            The convention is North=0.0, East=90.0, South=180.0, West=270.0.
            If wind is calm for the given hour, the direction equals zero.
            Values can range from 0 to 360. Missing value is 999.
        direct_normal_rad: The hourly Direct Normal Radiation in Wh/m2.
            Direc Normal Radiation is the amount of solar radiation in Wh/m2
            received directly from the solar disk on a surface perpendicular
            to the sun's rays. Missing values are (9999).
        diffuse_horizontal_rad: The hourly Diffuse Horizontal Radiation in Wh/m2.
            Diffuse Horizontal Radiation is the amount of solar radiation in
            Wh/m2 received from the sky (excluding the solar disk) on a
            horizontal surface. Missing values are (9999).
        global_horizontal_rad: The hourly Global Horizontal Radiation in Wh/m2.
            Global Horizontal Radiation is the total amount of direct and
            diffuse solar radiation in Wh/m2 received on a horizontal surface.
            It is not currently used inbEnergyPlus calculations. It should
            have a minimum value of 0; missing value for this field is 9999.
        horizontal_infrared_rad: The Horizontal Infrared Radiation Intensity in Wh/m2.
            If it is missing, EnergyPlus calculates it from the Opaque Sky Cover
            field. It should have a minimum value of 0; missing value is 9999.
        direct_normal_ill: The hourly Direct Normal Illuminance in lux.
            Direct Normal Illuminance is the average amount of illuminance in
            lux received directly from the solar disk on a surface
            perpendicular to the sun's rays. It is not currently used in
            EnergyPlus calculations. It should have a minimum value of 0;
            missing value is 999999.
        diffuse_horizontal_ill: The hourly Diffuse Horizontal Illuminance in lux.
            Diffuse Horizontal Illuminance is the average amount of illuminance
            in lux received from the sky (excluding the solar disk) on a
            horizontal surface. It is not currently used in EnergyPlus
            calculations. It should have a minimum value of 0; missing
            value is 999999.
        global_horizontal_ill: The hourly Global Horizontal Illuminance in lux.
            Global Horizontal Illuminance is the average total amount of
            direct and diffuse illuminance in lux received on a horizontal
            surface. It is not currently used in EnergyPlus calculations.
            It should have a minimum value of 0; missing value is 999999.
        total_sky_cover: The fraction for Total Sky Cover  in tenths of coverage.
            (eg. 1 is 1/10 covered. 10 is total coverage). Total Sky Cover is
            the amount of the sky dome covered by clouds or obscuring phenomena.
            Minium value is 0; maximum value is 10; missing value is 99.
        barometric_pressure: The hourly weather station pressure in Pa.
            Valid values range from 31,000 to 120,000. Missing value is 999999.
        model_year: The year from which the hourly data has been extracted.
            EPW files are synthesized from real recorded data from different
            years in a given climate. This is done to ensure that, for each
            month, the selected data is statistically representative of the
            average monthly conditions over the 18+ years of recording the data.
            Different EPW files will be synthesized from different years
            depeding on whether they are TMY (Typical Meteorological Year),
            TMY2, TMY3, AMY (Actual Meteorological Year) or other.
        ground_temperature: Monthly ground temperature data if it exists within
            the EPW file. Typically, each data collection in this list
            represents monthly temperatures at three different depths.
            - 0.5 meters
            - 2.0 meters
            - 4.0 meters
"""

ghenv.Component.Name = 'LB Import EPW'
ghenv.Component.NickName = 'ImportEPW'
ghenv.Component.Message = '1.8.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '0 :: Import'
ghenv.Component.AdditionalHelpFromDocStrings = '2'

try:
    import ladybug.epw as epw
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    epw_data = epw.EPW(_epw_file)
    location = epw_data.location
    dry_bulb_temperature = epw_data.dry_bulb_temperature
    dew_point_temperature = epw_data.dew_point_temperature
    relative_humidity = epw_data.relative_humidity
    wind_speed = epw_data.wind_speed
    wind_direction = epw_data.wind_direction
    direct_normal_rad = epw_data.direct_normal_radiation
    diffuse_horizontal_rad = epw_data.diffuse_horizontal_radiation
    global_horizontal_rad = epw_data.global_horizontal_radiation
    horizontal_infrared_rad = epw_data.horizontal_infrared_radiation_intensity
    direct_normal_ill = epw_data.direct_normal_illuminance
    diffuse_horizontal_ill = epw_data.diffuse_horizontal_illuminance
    global_horizontal_ill = epw_data.global_horizontal_illuminance
    total_sky_cover = epw_data.total_sky_cover
    barometric_pressure = epw_data.atmospheric_station_pressure
    model_year = epw_data.years
    g_temp = epw_data.monthly_ground_temperature
    ground_temperature = [g_temp[key] for key in sorted(g_temp.keys())]



# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2024, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Drop suggested Ladybug Tools components into a Grasshopper file for every
Legacy Ladybug + Honeybee component on the canvas.
-
All existing LBT and native Grasshopper components will be left as they are and
only the Legacy components will be circled in Red and have the suggested LBT
component placed next to them (if applicable). Note that, after this component
runs, you must then connect the new LBT components to the others and delete the
Legacy components.
-
Where applicable, each red circle will have a message about how the LBT component
differs from the Legacy one or if there may be a more appropirate LBT component
in the future. Also note that some Legacy workflows have been heavily refactored
since Legacy, meaning a different number of components may be necessary to achieve
the same thing (typically fewer in LBT than Legacy, meaning some LEgacy components
should be deleted without replacement).
-

    Args:
        _update: Set to "True" to have this component to search through the current
            Grasshopper file and drop suggested Ladybug Tools components
            for every Legacy Ladybug + Honeybee component on the canvas.

    Returns:
        report: Errors, warnings, etc.
"""

ghenv.Component.Name = 'LB Legacy Updater'
ghenv.Component.NickName = 'Legacy'
ghenv.Component.Message = '1.8.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '5 :: Version'
ghenv.Component.AdditionalHelpFromDocStrings = '0'

try:
    from ladybug_rhino.versioning.gather import gather_canvas_components
    from ladybug_rhino.versioning.legacy import suggest_new_component
    from ladybug_rhino.grasshopper import all_required_inputs, give_warning
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component) and _update:
    # load all of the GHPython userobjects and update the versions
    components = gather_canvas_components(ghenv.Component)
    report_init = []
    for comp in components:
        try:
            report_init.append(suggest_new_component(comp, ghenv.Component))
        except Exception:
            if hasattr(comp, 'Name'):
                msg = 'Failed to Update "{}"'.format(comp.Name)
                print(msg)
                give_warning(ghenv.Component, msg)
    print('\n'.join(r for r in report_init if r))



# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2024, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Calculate the percentage of people dissatisfied from radiant asymmetry.
_
The comfort functions used here come from Figure 5.2.4.1 of ASHRAE 55 2010.
Note that, if the resulting input results in a PPD beyond what is included
in this Figure, the maximum PPD will simply be returned.
-

    Args:
        _radiant_diff: A number for the the radiant temperature difference between
            two sides of the same plane where an occupant is located [C].
            This can also be a data collection representing the radiant
            temperature difference over time [C].
        _asymmetry_type: Text or an integer that representing the type of radiant
            asymmetry being evaluated. Occupants are more sensitive to warm
            ceilings and cool walls than cool ceilings and warm walls.
            Choose from the following options.
            _
            * 0 = WarmCeiling
            * 1 = CoolWall
            * 2 = CoolCeiling
            * 3 = WarmWall

    Returns:
        ppd: The percentage of people dissatisfied (PPD) for the input radiant asymmetry.
"""

ghenv.Component.Name = 'LB Radiant Asymmetry'
ghenv.Component.NickName = 'RadAsymm'
ghenv.Component.Message = '1.8.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '1 :: Analyze Data'
ghenv.Component.AdditionalHelpFromDocStrings = '0'

try:
    from ladybug_comfort.local import radiant_asymmetry_ppd
    from ladybug.datacollection import HourlyContinuousCollection
    from ladybug.datatype.fraction import PercentagePeopleDissatisfied
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))

ASYMMETRY_TYPES = {
    '0': 0,
    '1': 1,
    '2': 2,
    '3': 3,
    'WarmCeiling': 0,
    'CoolWall': 1,
    'CoolCeiling': 2,
    'WarmWall': 3
}



if all_required_inputs(ghenv.Component):
    ppd = HourlyContinuousCollection.compute_function_aligned(
        radiant_asymmetry_ppd, [_radiant_diff, ASYMMETRY_TYPES[_asymmetry_type]],
        PercentagePeopleDissatisfied(), '%')



# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2024, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Draw thermal comfort polygons on a Psychrometric Chart using the UTCI outdoor
thermal comfort model.
-

    Args:
        _psyc_chart: A ladybug Psychrometric Chart object on which the UTCI comfort
            polygons will be plot. This can be obtained from the "LB Psychrometric
            Chart" component.
        _mrt_: A number or list of numbers for the mean radiant temperature. These
            should be in Celsius if the Psychrometric Chart is in SI and
            Farenheit if the Psychrometric Chart is in IP. If None, a polygon for
            operative temperature will be plot, assuming that radiant temperature
            and air temperature are the same. (Default: None).
        _wind_speed_: A number or list of numbers for for the meteorological wind speed
            values in m/s (measured 10 m above the ground). If None, this will
            default to a low wind speed of 0.5 m/s, which is the lowest input
            speed that is recommended for the UTCI model.
        utci_par_: Optional UTCIParameter object to specify parameters under
            which conditions are considered acceptable. If None, default will
            assume comfort thresholds consistent with those used by meteorologists
            to categorize outdoor conditions.
        merge_poly_: Boolean to note whether all comfort polygons should be merged
            into a single polygon instead of separate polygons for each set of
            input conditions. (Default: False).
        plot_stress_: Boolean to note whether polygons for heat/cold stress should be plotted
            in the output. This will include 3 polygons on either side of the comfort
            polygon(s) for...
            _
            * Moderate Stress
            * Strong Stress
            * Very Strong Stress

    Returns:
        report: ...
        total_comfort: The percent of the data on the psychrometric chart that
            are inside all comfort polygons.
        total_comf_data: Data collection or a 0/1 value noting whether each of the data
            points on the psychrometric chart lies inside of a comfort polygon.
            _
            This can be connected to the "LB Create Legend" component to generate
            a list of colors that can be used to color the points output from
            "LB Psychrometric Chart" component to see exactly which points are
            comfortable and which are not.
            _
            Values are one of the following:
                0 = uncomfortable
                1 = comfortable
        polygons: A list of Breps representing the range of comfort (or heat/cold stress) for the
            input mrt and air speed.
        polygon_names:  A list of names for each of the polygons which correspond with the polygons
            output above. This will include both the comfort polygons and the
            cold/heat stress polygons.
        polygon_data: A list of data collections or 0/1 values indicating whether each 
            of the data points on the psychrometric chart lies inside each of
            the comfort polygons. Each data collection or here corresponds to the
            names in the polygon_names output above.
            _
            Values are one of the following:
                0 = outside
                1 = inside
"""

ghenv.Component.Name = 'LB UTCI Polygon'
ghenv.Component.NickName = 'UTCI Polygon'
ghenv.Component.Message = '1.8.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '2 :: Visualize Data'
ghenv.Component.AdditionalHelpFromDocStrings = '2'

try:
    from ladybug.psychchart import PsychrometricChart
    from ladybug.datacollection import BaseCollection
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_comfort.chart.polygonutci import PolygonUTCI
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_comfort:\n\t{}'.format(e))

try:
    from ladybug_rhino.config import tolerance
    from ladybug_rhino.fromgeometry import from_polyline2d_to_offset_brep
    from ladybug_rhino.grasshopper import all_required_inputs, \
        list_to_data_tree, give_warning
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


def strategy_warning(polygon_name):
    """Give a warning about a polygon not fitting on the chart."""
    msg = 'Polygon "{}" could not fit on the chart given the current location of ' \
        'the comfort polygon(s).\nTry moving the comfort polygon(s) by changing ' \
        'its criteria to see the missing polygon.'.format(polygon_name)
    give_warning(ghenv.Component, msg)
    print(msg)


def process_polygon(polygon_name, polygon):
    """Process a strategy polygon that does not require any special treatment."""
    if polygon is not None:
        polygon_names.append(polygon_name)
        strategy_poly.append(from_polyline2d_to_offset_brep(polygon, offset, z))
        dat = poly_obj.evaluate_polygon(polygon, tolerance)
        dat = dat[0] if len(dat) == 1 else poly_obj.create_collection(dat, polygon_name)
        polygon_data.append(dat)
    else:
        strategy_warning(polygon_name)


def merge_polygon_data(poly_data):
    """Merge an array of polygon comfort conditions into a single data list."""
    val_mtx = [dat.values for dat in poly_data]
    merged_values = []
    for hr_data in zip(*val_mtx):
        hr_val = 1 if 1 in hr_data else 0
        merged_values.append(hr_val)
    return merged_values


if all_required_inputs(ghenv.Component):
    # convert the temperature values to C if the polygon is in IP
    assert isinstance(_psych_chart, PsychrometricChart), 'PolygonUTCI ' \
        'psychrometric chart must be a ladybug PsychrometricChart. ' \
        'Got {}.'.format(type(_psych_chart))
    z = _psych_chart.z
    offset = _psych_chart.x_dim * 0.25
    if _psych_chart.use_ip:
        _mrt_ = PolygonUTCI.TEMP_TYPE.to_unit(_mrt_, 'C', 'F')

    # create the PolygonUTCI object
    poly_obj = PolygonUTCI(_psych_chart, _mrt_, _wind_speed_, comfort_parameter=utci_par_)

    # draw the comfort polygon
    polygon_names = []
    comfort_data = []
    if merge_poly_:
        polygons = [from_polyline2d_to_offset_brep(poly_obj.merged_comfort_polygon, offset, z)]
        comfort_data.append(poly_obj.merged_comfort_data)
        polygon_names.append('Comfort')
    else:
        polygons = [
            from_polyline2d_to_offset_brep(poly, offset, z)
            for poly in poly_obj.comfort_polygons
        ]
        comfort_data.extend(poly_obj.comfort_data)
        if len(polygons) == 1:
            polygon_names.append('Comfort')
        else:
            names = ('Comfort {}'.format(i + 1) for i in range(len(polygons)))
            polygon_names.extend(names)

    # draw the cold/heat stress polygons if requested
    polygon_data = comfort_data[:]
    if plot_stress_:
        stress_polys = (
            poly_obj.moderate_cold_polygon, poly_obj.strong_cold_polygon,
            poly_obj.very_strong_cold_polygon, poly_obj.moderate_heat_polygon,
            poly_obj.strong_heat_polygon, poly_obj.very_strong_heat_polygon
        )
        stress_names = (
            'Moderate Cold Stress', 'Strong Cold Stress',
            'Very Strong Cold Stress', 'Moderate Heat Stress',
            'Strong Heat Stress', 'Very Strong Heat Stress'
        )
        for poly, name in zip(stress_polys, stress_names):
            stress_data = poly_obj.evaluate_inside(poly[0], poly[2], name)
            if 'Cold' in name:
                polygons.insert(0, from_polyline2d_to_offset_brep(poly, offset, z))
                polygon_names.insert(0, name)
                polygon_data.insert(0, stress_data)
            else:
                polygons.append(from_polyline2d_to_offset_brep(poly, offset, z))
                polygon_names.append(name)
                polygon_data.append(stress_data)

    # compute total comfort values
    polygon_comfort = [dat.average * 100 for dat in comfort_data] if \
        isinstance(comfort_data[0], BaseCollection) else \
        [dat * 100 for dat in comfort_data]
    if isinstance(comfort_data[0], BaseCollection):
        merged_vals = merge_polygon_data(comfort_data)
        total_comf_data = poly_obj.create_collection(merged_vals, 'Total Comfort')
        total_comfort = total_comf_data.average * 100
    else:
        total_comf_data = 1 if sum(comfort_data) > 0 else 0
        total_comfort = total_comf_data * 100



# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2024, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Filter or select faces of geometry based on their orientation.

    Args:
        north_: A number between -360 and 360 for the counterclockwise difference between
            the North and the positive Y-axis in degrees. 90 is West and 270
            is East. This can also be Vector for the direction to North. (Default: 0)
        _geometry: Rhino Breps and/or Rhino Meshes which will be broken down into individual
            planar faces and filtered based on the direction they face.
        _orientation: Text for the direction that the geometry is facing. This can also be
            a number between 0 and 360 for the azimuth (clockwise horizontal
            degrees from North) that the geometry should face. Choose from
            the following:
            _
            * N = North
            * NE = Northeast
            * E = East
            * SE = Southeast
            * S = South
            * SW = Southwest
            * W = West
            * NW = Northwest
            * Up = Upwards
            * Down = Downwards
        _up_angle_: A number in degrees for the maximum declination angle from the positive
            Z Axis that is considerd up. This should be between 0 and 90 for
            the results to be practical. (Default: 30).
        _down_angle_: A number in degrees for the maximum angle difference from the newative
            Z Axis that is considerd down. This should be between 0 and 90 for
            the results to be practical. (Default: 30).
        _horiz_angle_: Angle in degrees for the horizontal deviation from _orientation
            that is still considered to face that orientation. This should be
            between 0 and 90 for the results to be practical. Note that this input
            has no effect when the input orientation is "Up" or "Down". (Default: 23).

    Returns:
        report: ...
        sel_geo: Selected faces of the input geometry that are facing the direction
            corresponding to the input criteria.
"""

ghenv.Component.Name = 'LB Filter by Normal'
ghenv.Component.NickName = 'FilterNormal'
ghenv.Component.Message = '1.8.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '4 :: Extra'
ghenv.Component.AdditionalHelpFromDocStrings = '0'

import math

try:
    from ladybug_geometry.geometry2d import Vector2D
    from ladybug_geometry.geometry3d import Vector3D
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_geometry:\n\t{}'.format(e))

try:
    from ladybug_rhino.togeometry import to_face3d, to_vector2d
    from ladybug_rhino.fromgeometry import from_face3d
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))

ORIENT_MAP = {
    'N': 0,
    'NE': 45,
    'E': 90,
    'SE': 135,
    'S': 180,
    'SW': 225,
    'W': 270,
    'NW': 315,
    'NORTH': 0,
    'NORTHEAST': 45,
    'EAST': 90,
    'SOUTHEAST': 135,
    'SOUTH': 180,
    'SOUTHWEST': 225,
    'WEST': 270,
    'NORTHWEST': 315,
    'UP': 'UP',
    'DOWN': 'DOWN',
    'UPWARDS': 'UP',
    'DOWNWARDS': 'DOWN'
}


if all_required_inputs(ghenv.Component):
    # process all of the global inputs
    if north_ is not None:  # process the north_
        try:
            north_ = to_vector2d(north_).angle_clockwise(Vector2D(0, 1))
        except AttributeError:  # north angle instead of vector
            north_ = math.radians(float(north_))
    else:
        north_ = 0
    up_angle = math.radians(_up_angle_) if _up_angle_ is not None else math.radians(30)
    down_angle = math.radians(_down_angle_) if _down_angle_ is not None else math.radians(30)
    horiz_angle =  math.radians(_horiz_angle_) if _horiz_angle_ is not None else math.radians(23)
    up_vec, down_vec = Vector3D(0, 0, 1), Vector3D(0, 0, -1)

    # process the geometry and the orientation
    all_geo = [f for geo in _geometry for f in to_face3d(geo)]
    try:
        orient = ORIENT_MAP[_orientation.upper()]
    except KeyError:
        try:
            orient = float(_orientation)
        except Exception:
            msg = 'Orientation must be text (eg. N, E, S W) or a number for the\n' \
                'azimuth of the geometry. Got {}.'.format(_orientation)
            raise TypeError(msg)

    # filter the geometry by the orientation
    if orient == 'UP':
        sel_geo = [f for f in all_geo if f.normal.angle(up_vec) < up_angle]
    elif orient == 'DOWN':
        sel_geo = [f for f in all_geo if f.normal.angle(down_vec) < down_angle]
    else:
        sel_geo = []
        dir_vec = Vector2D(0, 1).rotate(north_).rotate(-math.radians(orient))
        full_down_ang = math.pi - down_angle
        for f in all_geo:
            if up_angle <= f.normal.angle(up_vec) <= full_down_ang:
                norm_2d = Vector2D(f.normal.x, f.normal.y)
                if -horiz_angle <= norm_2d.angle(dir_vec) <= horiz_angle:
                    sel_geo.append(f)

    # translate the Face3D back to Rhino geometry
    sel_geo = [from_face3d(f) for f in sel_geo]



# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2024, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Create a colored plot of any hourly data collection.
-

    Args:
        _data: A HourlyContinuousCollection or HourlyDiscontinuousCollection
            which will be used to generate the hourly plot.
        _base_pt_: An optional Point3D to be used as a starting point to generate
            the geometry of the plot (Default: (0, 0, 0)).
        _x_dim_: A number to set the X dimension of the mesh cells (Default: 1 meters).
        _y_dim_: A number to set the Y dimension of the mesh cells (Default: 4 meters).
        _z_dim_: A number to set the Z dimension of the entire chart. This will
            be used to make the colored_mesh3d of the chart vary in the Z
            dimension according to the data. The value input here should usually be
            several times larger than the x_dim or y_dim in order to be noticable
            (e.g. 100). If 0, the colored_mesh3d will simply be flat. (Default: 0).
        reverse_y_: Boolean to note whether the Y-axis of the chart is reversed
            If True, time over the course of the day will flow from the top of
            the chart to the bottom instead of the bottom to the top.
        clock_24_: Boolean to note whether the hour labels on the Y-Axis of the chart
            should be in 24-hour clock format (eg. 18:00) or they should be
            in 12-hour clock format (eg. 6PM).
        legend_par_: An optional LegendParameter object to change the display of the
            HourlyPlot. This can also be a list of legend parameters to be
            applied to the different connected _data.
        statement_: A conditional statement as a string (e.g. a > 25).
            .
            The variable of the first data collection should always be named 'a'
            (without quotations), the variable of the second list should be
            named 'b', and so on.
            .
            For example, if three data collections are connected to _data
            and the following statement is applied:
            '18 < a < 26 and b < 80 and c > 2'
            The resulting collections will only include values where the first
            data collection is between 18 and 26, the second collection is less
            than 80 and the third collection is greater than 2.
        period_: A Ladybug analysis period to be applied to all of the input _data.

    Returns:
        report: ...
        mesh: A colored mesh derived from the input _data. Multiple meshes will
            be output for several data collections are input.
        legend: Geometry representing the legend for each mesh.
        borders: A list of lines and polylines representing different time
            intervals of the plot.
        labels: A list of text objects that label the borders with the time
            intervals that they demarcate.
        title: A text object for the global_title.
        vis_set: An object containing VisualizationSet arguments for drawing a detailed
            version of the Hourly Plot in the Rhino scene. This can be connected to
            the "LB Preview Visualization Set" component to display this version
            of the Hourly Plot in Rhino.
"""

ghenv.Component.Name = "LB Hourly Plot"
ghenv.Component.NickName = 'HourlyPlot'
ghenv.Component.Message = '1.8.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '2 :: Visualize Data'
ghenv.Component.AdditionalHelpFromDocStrings = '1'

try:
    from ladybug.datacollection import HourlyContinuousCollection
    from ladybug.hourlyplot import HourlyPlot
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_geometry.geometry3d.pointvector import Point3D
    from ladybug_geometry.geometry3d.plane import Plane
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_geometry:\n\t{}'.format(e))

try:
    from ladybug_rhino.config import conversion_to_meters
    from ladybug_rhino.togeometry import to_point3d
    from ladybug_rhino.fromgeometry import from_mesh3d, from_mesh2d, \
        from_polyline2d, from_linesegment2d
    from ladybug_rhino.text import text_objects
    from ladybug_rhino.fromobjects import legend_objects
    from ladybug_rhino.grasshopper import all_required_inputs, list_to_data_tree, \
        objectify_output
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    # apply any analysis periods and conditional statements to the input collections
    if period_ is not None:
        _data = [coll.filter_by_analysis_period(period_) for coll in _data]
    if statement_ is not None:
        _data = HourlyContinuousCollection.filter_collections_by_statement(
            _data, statement_)

    # set default values for the chart dimensions
    _base_pt_ = to_point3d(_base_pt_) if _base_pt_ is not None else Point3D()
    _x_dim_ = _x_dim_ if _x_dim_ is not None else 1.0 / conversion_to_meters()
    _y_dim_ = _y_dim_ if _y_dim_ is not None else 4.0 / conversion_to_meters()
    _z_dim_ = _z_dim_ if _z_dim_ is not None else 0
    reverse_y_ = reverse_y_ if reverse_y_ is not None else False

    # empty lists of objects to be filled with visuals
    mesh, title, all_legends, all_borders, all_labels, vis_set = [], [], [], [], [], []

    for i, data_coll in enumerate(_data):
        try:  # sense when several legend parameters are connected
            lpar = legend_par_[i]
        except IndexError:
            lpar = None if len(legend_par_) == 0 else legend_par_[-1].duplicate()

        # create the hourly plot object and get the main pieces of geometry
        hour_plot = HourlyPlot(data_coll, lpar, _base_pt_,
                               _x_dim_, _y_dim_, _z_dim_, reverse_y_)
        
        msh = from_mesh2d(hour_plot.colored_mesh2d, _base_pt_.z) if _z_dim_ == 0 else \
            from_mesh3d(hour_plot.colored_mesh3d)
        mesh.append(msh)
        border = [from_polyline2d(hour_plot.chart_border2d, _base_pt_.z)] + \
            [from_linesegment2d(line, _base_pt_.z) for line in hour_plot.hour_lines2d] + \
            [from_linesegment2d(line, _base_pt_.z) for line in hour_plot.month_lines2d]
        all_borders.append(border)
        legnd = legend_objects(hour_plot.legend)
        all_legends.append(legnd)
        tit_txt = text_objects(hour_plot.title_text, hour_plot.lower_title_location,
                               hour_plot.legend_parameters.text_height,
                               hour_plot.legend_parameters.font)
        title.append(tit_txt)

        # create the text label objects
        hr_text = hour_plot.hour_labels_24 if clock_24_ else hour_plot.hour_labels
        label1 = [text_objects(txt, Plane(o=Point3D(pt.x, pt.y, _base_pt_.z)),
                               hour_plot.legend_parameters.text_height,
                               hour_plot.legend_parameters.font, 2, 3)
                  for txt, pt in zip(hr_text, hour_plot.hour_label_points2d)]
        label2 = [text_objects(txt, Plane(o=Point3D(pt.x, pt.y, _base_pt_.z)),
                               hour_plot.legend_parameters.text_height,
                               hour_plot.legend_parameters.font, 1, 0)
                  for txt, pt in zip(hour_plot.month_labels, hour_plot.month_label_points2d)]
        all_labels.append(label1 + label2)

        # increment the base point so that the next chart doesn't overlap this one
        try:
            next_aper = _data[i + 1].header.analysis_period
            next_tstep = next_aper.timestep
            next_hour = next_aper.end_hour - next_aper.st_hour + 1
        except IndexError:
            next_tstep = 1
            next_hour = 24
        txt_dist = hour_plot.legend_parameters.text_height * (len(_data[i].header.metadata) + 6) * 1.5
        increment = (next_hour * next_tstep * _y_dim_) + txt_dist
        _base_pt_ = Point3D(_base_pt_.x, _base_pt_.y - increment, _base_pt_.z)

        # append the VisualizationSet arguments with fixed geometry
        hp_leg_par3d = hour_plot.legend_parameters.properties_3d
        hp_leg_par3d.base_plane = hp_leg_par3d.base_plane
        hp_leg_par3d.segment_height = hp_leg_par3d.segment_height
        hp_leg_par3d.segment_width = hp_leg_par3d.segment_width
        hp_leg_par3d.text_height = hp_leg_par3d.text_height
        vis_set.append((hour_plot, _base_pt_.z))

    # convert nexted lists into data trees
    legend = list_to_data_tree(all_legends)
    borders = list_to_data_tree(all_borders)
    labels = list_to_data_tree(all_labels)
    vis_set = objectify_output('VisualizationSet Aruments [HourlyPlot]', vis_set)



# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2024, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Compute a True North angle and vector from Magnetic North at a given location.
_
This component uses then World Magnetic Model (WMM) developed and maintained by NOAA.
https://www.ncei.noaa.gov/products/world-magnetic-model
-

    Args:
        _location: A ladybug Location that has been output from the "LB Import EPW"
            component or the "LB Construct Location" component. This is used
            to determine the difference between magnetic and true North.
        _mag_north: A number between -360 and 360 for the counterclockwise difference
            between Magnetic North and the positive Y-axis in degrees. Counterclockwise
            means "90 is West and 270 is East". This can also be Vector for the
            magnetic North direction.
        _year_: A number for the year in which the Magnetic North was evaluated. Decimal
            values are accepted. This is needed as the location of Magnetic
            North has been moving at a rate of roughly 50 km/year for the
            past couple of decades. (Default: 2025).
        cof_file_: An optional path to a .COF file containing the coefficients that form the
            inputs for the World Magnetic Model (WMM). A new set of coefficients
            is published roughly every 5 years as the magnetic poles continue to
            move. If unspecified, coefficients will be taken from the most
            recent model. COF files with the most recent coefficients and
            historical values are available at:
            https://www.ncei.noaa.gov/products/world-magnetic-model/wmm-coefficients

    Returns:
        mag_declination: The magnetic declination in degrees. Magnetic declination is the
        difference between magnetic North and true North at a given location on
        the globe (expressed in terms of degrees).
        true_north: A number between -360 and 360 for the True North angle in degrees.
        true_north_vec: A vector for the True North direction. This can be plugged into
            any of the north_ inputs of the other LAdybug Tools components.
"""

ghenv.Component.Name = 'LB Magnetic to True North'
ghenv.Component.NickName = 'TrueNorth'
ghenv.Component.Message = '1.8.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '4 :: Extra'
ghenv.Component.AdditionalHelpFromDocStrings = '6'

import math

try:
    from ladybug_geometry.geometry2d.pointvector import Vector2D
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_geometry:\n\t{}'.format(e))

try:
    from ladybug.north import WorldMagneticModel
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_geometry:\n\t{}'.format(e))

try:
    from ladybug_rhino.togeometry import to_vector2d
    from ladybug_rhino.fromgeometry import from_vector2d
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    # process the _magn_north and the year
    try:
        _mag_north = math.degrees(to_vector2d(_mag_north).angle_clockwise(Vector2D(0, 1)))
    except AttributeError:  # north angle instead of vector
        _mag_north = float(_mag_north)
    _year_ = _year_ if _year_ is not None else 2025

    # initialize the WorldMagneticModel and convert the north angle
    wmm_obj = WorldMagneticModel(cof_file_)
    true_north = wmm_obj.magnetic_to_true_north(_location, _mag_north, _year_)
    mag_declination = true_north - _mag_north

    # convert the true north angle to a vector
    st_north = Vector2D(0, 1).rotate(math.radians(_mag_north))
    true_north_vec = st_north.rotate(math.radians(true_north))
    true_north_vec = from_vector2d(true_north_vec)



# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2024, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Calculate humidity metrics from relative humidity, dry bulb temperature and
(if present) atmospheric pressure.
-

    Args:
        _dry_bulb: A value or data collection representing  dry bulb temperature [C]
        _rel_humid: A value or data collection representing relative humidity [%]
        _pressure_: A value or data collection representing atmospheric pressure [Pa]
            Default is to use air pressure at sea level (101,325 Pa).
    
    Returns:
        humid_ratio: A data collection or value for humidity ratio
            (aka. absolute humidity). Units are fractional (kg water / kg air).
        enthalpy: A data collection or value for enthalpy (kJ / Kg).
        wet_bulb: A data collection or value for wet bulb temperature (C).
        dew_point: A data collection or value for dew point temperature (C).
"""

ghenv.Component.Name = "LB Humidity Metrics"
ghenv.Component.NickName = 'HumidityR'
ghenv.Component.Message = '1.8.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '1 :: Analyze Data'
ghenv.Component.AdditionalHelpFromDocStrings = '0'

try:
    from ladybug.psychrometrics import humid_ratio_from_db_rh, enthalpy_from_db_hr, \
        wet_bulb_from_db_rh, dew_point_from_db_rh
    from ladybug.datacollection import HourlyContinuousCollection
    from ladybug.datatype.fraction import HumidityRatio
    from ladybug.datatype.specificenergy import Enthalpy
    from ladybug.datatype.temperature import WetBulbTemperature, DewPointTemperature
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    if _pressure_ is None:
        _pressure_ = 101325

    humid_ratio = HourlyContinuousCollection.compute_function_aligned(
        humid_ratio_from_db_rh, [_dry_bulb, _rel_humid, _pressure_],
        HumidityRatio(), 'fraction')
    
    enthalpy = HourlyContinuousCollection.compute_function_aligned(
        enthalpy_from_db_hr, [_dry_bulb, humid_ratio], Enthalpy(), 'kJ/kg')

    wet_bulb = HourlyContinuousCollection.compute_function_aligned(
        wet_bulb_from_db_rh, [_dry_bulb, _rel_humid, _pressure_],
        WetBulbTemperature(), 'C')

    dew_point = HourlyContinuousCollection.compute_function_aligned(
        dew_point_from_db_rh, [_dry_bulb, _rel_humid],
        DewPointTemperature(), 'C')


# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2024, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Deconstruct a Ladybug VisualizationSet into all of its constituent objects.
_
This includes Context Geometry, Analysis Geometry, and any data sets that are
associated with the analysis geometry. The last one is particularly helpful
for performing analysis in the data associated with a particular visualization.
-

    Args:
        _vis_set: VisualizationSet arguments from any Ladybug Tools component with a vis_set
            output. This can also be the path to a .vsf file that exists on this
            machine (these files are often written with the "LB Dump VisualizationSet"
            component). Lastly, this input can be a custom VisualizationSet that
            has been created with the Ladybug Tools SDK.

    Returns:
        context: A list of geometry objects that constitute the context geometry of the
            VisualizationSet. When the VisualizationSet contains multiple context
            geometry instances, this will be a data tree with one branch for
            each context object.
        analysis: A list of geometry objects that constitute the analysis geometry of the
            VisualizationSet. When the VisualizationSet contains multiple analysis
            geometry instances, this will be a data tree with one branch for
            each analysis object.
        data: A list of numbers that constitue the data set associated with the analysis
            geometry. In the event of multiple data sets assigned to the same
            analysis geometry, this will be a data tree of numbers with one branch
            for each data set. In the event of multiple analysis geometries, this
            will be a nested data tree where the first number in the path matches
            the analysis geometry branch and the last number matches the data set
            number.
"""

ghenv.Component.Name = "LB Deconstruct VisualizationSet"
ghenv.Component.NickName = 'DeconstructVisSet'
ghenv.Component.Message = '1.8.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '4 :: Extra'
ghenv.Component.AdditionalHelpFromDocStrings = '0'

try:  # import the honeybee dependencies
    from ladybug_display.geometry3d import DisplayText3D
    from ladybug_display.visualization import ContextGeometry, AnalysisGeometry
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_display:\n\t{}'.format(e))

try:  # import the ladybug_rhino dependencies
    from ladybug_rhino.fromobjects import from_geometry
    from ladybug_rhino.text import text_objects
    from ladybug_rhino.visset import process_vis_set
    from ladybug_rhino.grasshopper import all_required_inputs, list_to_data_tree
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))

TEXT_HORIZ = {'Left': 0, 'Center': 1, 'Right': 2}
TEXT_VERT = {'Top': 0, 'Middle': 3, 'Bottom': 5}


if all_required_inputs(ghenv.Component):
    # extract the VisualizationSet object
    _vs = process_vis_set(_vis_set)

    # loop through the constituient objects and deconstruct them
    context, analysis, data = [], [], []
    for geo_obj in _vs.geometry:
        if isinstance(geo_obj, ContextGeometry):
            con_geos = []
            for g in geo_obj.geometry:
                if not isinstance(g, DisplayText3D):
                    con_geos.append(from_geometry(g.geometry))
                else:
                    t_obj = text_objects(
                        g.text, g.plane, g.height, g.font,
                        TEXT_HORIZ[g.horizontal_alignment],
                        TEXT_VERT[g.vertical_alignment])
                    con_geos.append(t_obj)
            context.append(con_geos)
        elif isinstance(geo_obj, AnalysisGeometry):
            a_geos, data_sets = [], []
            for g in geo_obj.geometry:
                a_geos.append(from_geometry(g))
            for d in geo_obj.data_sets:
                data_sets.append(d.values)
            analysis.append([a_geos])
            data.append(data_sets)

    # convert everything into data trees
    context = list_to_data_tree(context)
    analysis = list_to_data_tree(analysis)
    data = list_to_data_tree(data)



# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2024, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>


"""
Create a set of parameters that define the acceptable conditions of the
Adaptive thermal comfort model.
-
These parameters can be plugged into any of the components that compute
Adaptive thermal comfort.
-

    Args:
        _ashrae_or_en_: A boolean to note whether to use the ASHRAE-55 neutral
            temperature function (True) or the european neutral function (False),
            which is consistent with EN-16798 (and the older EN-15251).
            Preference is given to EN-16798 in places where EN-16798 and
            EN-15251 differ (eg. the 1C lower cold threshold in EN-16798).
            Note that this input will also determine default values for
            many of the other properties of this object.
        _neutral_offset_:  The number of degrees Celcius from the neutral temperature
            where the input operative temperature is considered acceptable.
            The default is 2.5C when the neutral temperature function is ASHRAE-55
            (consistent with 90% PPD) and 3C when the neutral temperature
            function is EN-16798 (consistent with comfort class II). Note that,
            when the neutral temperature function is EN-16798, one degree
            Celsius is automatically added to the offset value input here to
            get the lower temperature threshold. This accounts for the fact that
            EN-16798 does not interpret the neutral offset symmetically.
            _
            For ASHRAE-55, the following neutral offsets apply.
                * 90 PPD - 2.5C
                * 80 PPD - 3.5C
            _
            For the EN standard, the following neutral offsets apply.
                * Class I - 2C
                * Class II - 3C
                * Class III - 4C
        _avgm_or_runmean_: A boolean to note whether the prevailing outdoor
            temperature is computed from the average monthly temperature (True) or
            a weighted running mean of the last week (False).  The default is True
            when the neutral temperature function is ASHRAE-55 and False when the
            neutral temperature function is EN.
        _discr_or_cont_vel_: A boolean to note whether discrete categories should be
            used to assess the effect of elevated air speed (True) or whether
            a continuous function should be used (False). Note that continuous
            air speeds were only used in the older EN-15251 standard and are
            not a part of the more recent EN-16798 standard.
            When unassigned, this will be True for discrete air speeds.
        _cold_prevail_limit_: A number indicating the prevailing outdoor temperature
            below which acceptable indoor operative temperatures flat line.
            The default is 10C, which is consistent with both ASHRAE-55 and
            EN-16798. However, 15C was used for the older EN-15251 standard.
            This number cannot be greater than 22C and cannot be less than 10C.
        _conditioning_: A number between 0 and 1 that represents how "conditioned"
            vs. "free-running" the building is.
                0 = free-running (completely passive with no air conditioning)
                1 = conditioned (no operable windows and fully air conditioned)
            The default is 0 since both the ASHRAE-55 and the EN standards prohibit
            the use of adaptive comfort methods when a heating/cooling system is active.
            When set to a non-zero number, a neutral temperature function for
            heated/cooled operation derived from the SCATs database will be used.
            For more information on how adaptive comfort methods can be applied to
            conditioned buildings, see the neutral_temperature_conditioned function
            in the ladybug_comfort documentation.

    Returns:
        adapt_par: An Adaptive comfort parameter object that can be plugged into
            any of the components that compute Adaptive thermal comfort.
"""

ghenv.Component.Name = 'LB Adaptive Comfort Parameters'
ghenv.Component.NickName = 'AdaptPar'
ghenv.Component.Message = '1.8.1'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '4 :: Extra'
ghenv.Component.AdditionalHelpFromDocStrings = '4'

try:
    from ladybug_comfort.parameter.adaptive import AdaptiveParameter
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_comfort:\n\t{}'.format(e))
try:
    from ladybug_rhino.grasshopper import turn_off_old_tag
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))
turn_off_old_tag(ghenv.Component)


adapt_par = AdaptiveParameter(_ashrae_or_en_, _neutral_offset_,
                              _avgm_or_runmean_, _discr_or_cont_vel_,
                              _cold_prevail_limit_, _conditioning_)



# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2024, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Select a sub-region of a mesh using aligned values and conditional criteria.
_
This has multiple uses and can be applied to any study that outputs a list of
results that are aligned with a mesh. For example, quantifying the daylit area
from a daylight analysis, selecting the portion of a roof with enough solar
radiation for photovoltaic panels, etc.
-

    Args:
        _values: A list of numbers that correspond to either the number of faces
            or vertices of the _mesh.
        _mesh: The mesh from which a sub-region will be selected. This is typically
            a colored mesh output from a study.
        _operator_: A text string representing an operator for the the conditional statement. 
            The default is greater than (>).  This must be an operator in python
            and examples include the following:
                * > - Greater Than
                * < - Less Than
                * >= - Greater or Equal
                * <= - Less or Equal
                * == - Equals
                * != - Does not Equal
        _pct_threshold_: A number between 0 and 100 that represents the percentage
            of the mesh faces or vertices to be included in the resulting
            sub_mesh. (Default: 25%).
        abs_threshold_: An optional number that represents the absolute threshold above
            which a given mesh face or vertex is included in the resulting
            sub_mesh. An input here will override the percent threshold input
            above.

    Returns:
        report: Reports, errors, warnings, etc.
        total_value: The sum of each value that meets the criteria multiplied by the
            corresponding mesh face area. This can generally be used to
            understand how much value is captured according to the conditional
            critera. For example, if the input _mesh is a radiation study,
            this is equal to the total radiation falling on the sub_mesh.
            This may or may not be meaningful depending on the units of the
            connected _values. This output will always be zero for cases
            where values correspond to mesh vertices and not faces.
        total_area: The area of the sub_mesh that meets the criteria.
        sub_mesh: A new mesh with the faces or vertices removed to reveal just the portion
            that satisfies the conditional criteria. By default, this is hidden
            to that just the outline appears in the geometry preview.
        outline: A set of lines outlining the portion of the mesh that is above
            the threshold.
"""

ghenv.Component.Name = 'LB Mesh Threshold Selector'
ghenv.Component.NickName = 'MeshSelector'
ghenv.Component.Message = '1.8.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '4 :: Extra'
ghenv.Component.AdditionalHelpFromDocStrings = '1'

try:
    from ladybug_rhino.togeometry import to_mesh3d
    from ladybug_rhino.fromgeometry import from_mesh3d_to_outline
    from ladybug_rhino.grasshopper import all_required_inputs, hide_output
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    # check the input values and provide defaults
    lb_mesh = to_mesh3d(_mesh)
    val_count = len(_values)
    face_match = val_count == len(lb_mesh.faces)
    assert face_match or val_count == len(lb_mesh.vertices), \
        'Number of _values ({}) must match the number of mesh faces ({}) or ' \
        'the number of mesh vertices ({}).'.format(
            val_count, len(lb_mesh.faces), len(lb_mesh.faces))
    fract_thresh = 0.25 if _pct_threshold_ is None else _pct_threshold_ / 100
    operator = '>' if _operator_ is None else _operator_.strip()
    if operator in ('==', '!='):
        assert abs_threshold_ is not None, 'An abs_threshold_ must be ' \
            'specified to use the "{}" operator.'.format(operator)

    # get a list of boolean values that meet the conditional criteria
    if abs_threshold_ is not None:
        statement = '{} ' + operator + str(abs_threshold_)
        pattern = []
        for val in _values:
            pattern.append(eval(statement.format(val), {}))
    else:
        pattern = [False] * val_count
        target_count = int(fract_thresh * (val_count))
        face_i_sort = [x for (y, x) in sorted(zip(_values, range(val_count)))]
        rel_values = list(reversed(face_i_sort)) if '>' in operator else face_i_sort
        for cnt in rel_values[:target_count]:
            pattern[cnt] = True

    # remove the faces or vertices from the mesh and compute the outputs
    total_value, total_area = 0, 0
    try:
        sub_mesh_lb, vf_pattern = lb_mesh.remove_faces(pattern) if face_match else \
            lb_mesh.remove_vertices(pattern)
        if face_match:
            for incl, val, area in zip(pattern, _values, lb_mesh.face_areas):
                if incl:
                    total_area += area
                    total_value += val * area
        else:
            total_area = sub_mesh_lb.area
        # convert everything to Rhino geometry
        sub_mesh, outline = from_mesh3d_to_outline(sub_mesh_lb)
        hide_output(ghenv.Component, 3)
    except AssertionError as e:
        if not 'Mesh must have at least one face' in str(e):
            raise AssertionError(e)


# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2024, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Calculate thermal indices that have historically been used by meteorologists.
All of them are "feels like" temperatures that attempt to account for factors
beyond sir temperature. These include the following:
_
* Wet Bulb Globe Temperature (WBGT)
* Heat Index (HI)
* Wind Chill Temperature (WCT)
_
Most of these indices have fallen out of use in favor of Universal Thermal
Climate Index (UTCI). However, they are still used in some regions and are a
part of older codes and standards.
-

    Args:
        _air_temp: Data Collection or individual value for air temperature in C.
            This input is used by all three metrics.
        _mrt_: Data Collection or individual value for mean radiant temperature
            (MRT) in C. Default is the same as the air_temp. This input only
            affects the WBGT.
        _rel_humid: Data Collection or individual value for relative humidity in %.
            Note that percent values are between 0 and 100. This input affects
            WBGT as well as HI.
        _wind_vel: Data Collection or individual value for meteoroligical wind velocity
            at 10 m above ground level in m/s. This is used by both WBGT and WCT.
    
    Returns:
        wbgt: A data collection or value for Wet Bulb Globe Temperature (WBGT) [C].
            WBGT is a type of feels-like temperature that is widely used as a
            heat stress index (ISO 7243). It is incorporates the effect of
            temperature, humidity, wind speed, and mean radiant temperature
            (optionally including the effect of sun).
        heat_index: A data collection or value for Heat Index (HI) temperature [C].
            Heat index is derived from original work carried out by Robert G.
            Steadman, which defined heat index through large tables of empirical
            data. The formula here approximates the heat index to within +/- 0.7C
            and is the result of a multivariate fit. Heat index was adopted by
            the US's National Weather Service (NWS) in 1979.
        wind_chill: A data collection or value for Wind Cill Temperature (WCT) [C].
            Wind Chill Index is derived from original work carried out by
            Gregorczuk. It qualifies thermal sensations of a person in
            wintertime. It is especially useful at low and very low air
            temperature and at high wind speed.
"""

ghenv.Component.Name = 'LB Thermal Indices'
ghenv.Component.NickName = 'ThermalIndices'
ghenv.Component.Message = '1.8.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '1 :: Analyze Data'
ghenv.Component.AdditionalHelpFromDocStrings = '0'

try:
    from ladybug_comfort.wbgt import wet_bulb_globe_temperature
    from ladybug_comfort.hi import heat_index as heat_index_temperature
    from ladybug_comfort.wc import windchill_temp
    from ladybug.datacollection import HourlyContinuousCollection
    from ladybug.datatype.temperature import WetBulbGlobeTemperature, \
        HeatIndexTemperature, WindChillTemperature
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    if _mrt_ is None:
        _mrt_ = _air_temp

    wbgt = HourlyContinuousCollection.compute_function_aligned(
        wet_bulb_globe_temperature, [_air_temp, _mrt_, _wind_vel, _rel_humid],
        WetBulbGlobeTemperature(), 'C')

    heat_index = HourlyContinuousCollection.compute_function_aligned(
        heat_index_temperature, [_air_temp, _rel_humid],
        HeatIndexTemperature(), 'C')

    wind_chill = HourlyContinuousCollection.compute_function_aligned(
        windchill_temp, [_air_temp, _wind_vel],
        WindChillTemperature(), 'C')


# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2024, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Get a matrix containing radiation values from each patch of a sky dome.
_
Creating this matrix is a necessary pre-step before doing incident radiation
analysis with Rhino geometry or generating a radiation rose.
_
This component uses Radiance's gendaymtx function to calculate the radiation
for each patch of the sky. Gendaymtx is written by Ian Ashdown and Greg Ward.
Morere information can be found in Radiance manual at:
http://www.radiance-online.org/learning/documentation/manual-pages/pdfs/gendaymtx.pdf
-

    Args:
        north_: A number between -360 and 360 for the counterclockwise
            difference between the North and the positive Y-axis in degrees.
            90 is West and 270 is East. This can also be Vector for the
            direction to North. (Default: 0)
        _location: A ladybug Location that has been output from the "LB Import EPW"
            component or the "LB Construct Location" component.
        _direct_rad: An annual hourly DataCollection of Direct Normal Radiation such
            as that which is output from the "LB Import EPW" component or the
            "LB Import STAT" component.
        _diffuse_rad: An annual hourly DataCollection of Diffuse Horizontal Radiation
            such as that which is output from the "LB Import EPW" component or
            the "LB Import STAT" component.
        _hoys_: A number or list of numbers between 0 and 8760 that respresent
            the hour(s) of the year for which to generate the sky matrix. The
            "LB Calculate HOY" component can output this number given a month,
            day and hour. The "LB Analysis Period" component can output a
            list of HOYs within a certain hour or date range. By default,
            the matrix will be for the entire year.
        high_density_: A Boolean to indicate whether the higher-density Reinhart sky
            matrix should be generated (True), which has roughly 4 times the sky
            patches as the (default) original Tregenza sky (False). Note that,
            while the Reinhart sky has a higher resolution and is more accurate,
            it will result in considerably longer calculation time for incident
            radiation studies. The difference in sky resolution can be observed
            with the "LB Sky Dome" component. (Default: False).
        _ground_ref_: A number between 0 and 1 to note the average ground reflectance
            that is associated with the sky matrix. (Default: 0.2).
        _folder_: The folder in which the Radiance commands are executed to
            produce the sky matrix. If None, it will be written to Ladybug's
            default EPW folder.

    Returns:
        report: ...
        sky_mtx: A sky matrix object containing the radiation coming from each patch
            of the sky. This can be used for a radiation study, a radition rose,
            or a sky dome visualization. It can also be deconstructed into its
            individual values with the "LB Deconstruct Matrix" component.
"""

ghenv.Component.Name = 'LB Cumulative Sky Matrix'
ghenv.Component.NickName = 'SkyMatrix'
ghenv.Component.Message = '1.8.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '2 :: Visualize Data'
ghenv.Component.AdditionalHelpFromDocStrings = '3'

import math

try:
    from ladybug_geometry.geometry2d.pointvector import Vector2D
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_geometry:\n\t{}'.format(e))

try:
    from ladybug_radiance.skymatrix import SkyMatrix
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_rhino.togeometry import to_vector2d
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))

try:
    from lbt_recipes.version import check_radiance_date
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_radiance:\n\t{}'.format(e))

# check the istalled Radiance date and get the path to the gemdaymtx executable
check_radiance_date()


if all_required_inputs(ghenv.Component):
    # process and set defaults for all of the global inputs
    if north_ is not None:  # process the north_
        try:
            north_ = math.degrees(
                to_vector2d(north_).angle_clockwise(Vector2D(0, 1)))
        except AttributeError:  # north angle instead of vector
            north_ = float(north_)
    else:
        north_ = 0
    ground_r = 0.2 if _ground_ref_ is None else _ground_ref_

    # create the sky matrix object
    sky_mtx = SkyMatrix.from_components(
        _location, _direct_rad, _diffuse_rad, _hoys_, north_, high_density_, ground_r)
    if _folder_:
        sky_mtx.folder = _folder_



# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2024, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Get a ray tracing visualization of direct sunlight rays reflected off of _source_geo
and subsequently bouncing through a set of context_ geometries.
_
Examples where this visualization could be useful include understading the
reflection of light by a light shelf or testing to see whether a parabolic
glass or metal building geometry might focus sunlight to dangerous levels at
certain times of the year.
_
Note that this component assumes that all sun light is reflected specularly
(like a mirror) and, for more detailed raytracing analysis with diffuse
scattering, the Honeybee Radiance components should be used.
-

    Args:
        _vector: A sun vector (typically from the "LB SunPath" component), which will
            be used to evaluate the light boucing off of the _source_geo and
            through the context_.
        _source_geo: A brep or mesh representing a surface off of which sun rays first
            bounce. Lists of breps or meshes are also acceptable. These surfaces
            will be used to generate the initial sun rays in a grid-like pattern.
        context_: Breps or meshes for conext geometry, which will reflect the sun rays
            after they bounce off of the _source_geo. They can also block the
            sun rays to the _source_geo if context_block_ is set to True.
        receiver_: Optional breps or meshes for receiver geometry for which sun ray
            intersection is being studied. If specified, only sun rays that
            reflect off of the _source_geo and have their last bounce intersect
            this receiver within the _last_length_ distance will appear in
            the result. This can help filter results to only an area of
            intersect where the impact of reflected sun rays is a concern.
        _grid_size: A positive number in Rhino model units for the average distance between
            sun ray points to generate along the _source_geo.
        _bounce_count_: An positive integer for the number of ray bounces to trace
            the sun rays forward. (Default: 1).
        _first_length_: A positive number in Rhino model units for the length of the
            sun ray before the first bounce. If unspecified, this will be
            the diagonal of the bounding box surrounding all input geometries.
        _last_length_: A positive number in Rhino model units representing the length
            of the sun ray after the last bounce. If unspecified, this will be
            the diagonal of the bounding box surrounding all input geometries.
        context_block_: Set to "True" to count the input _context as something that fully
            blocks rays as opposed to having the rays simply bounce off of it.
            This can be particularly useful in cases with lots of rays where
            you are only concered about the rays that can actually hit the
            _source_geo. (Default: False).
        _run: Set to "True" to run the component and perform the ray tracing study.

    Returns:
        report: Reports, errors, warnings, etc.
        rays: A list of polylines representing the sun rays traced forward onto
            the _source_geo and then through the context_.
        int_pts: A data tree of intersection points one one branch for each of the
            rays above.
"""

ghenv.Component.Name = 'LB Surface Ray Tracing'
ghenv.Component.NickName = 'SrfRayTrace'
ghenv.Component.Message = '1.8.1'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '3 :: Analyze Geometry'
ghenv.Component.AdditionalHelpFromDocStrings = '3'

import math

try:
    from ladybug_geometry.geometry3d.ray import Ray3D
    from ladybug_geometry.geometry3d.polyline import Polyline3D
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_geometry:\n\t{}'.format(e))

try:
    from ladybug_rhino.config import tolerance
    from ladybug_rhino.togeometry import to_joined_gridded_mesh3d, to_point3d, \
        to_vector3d
    from ladybug_rhino.fromgeometry import from_point3d, from_vector3d, from_ray3d, \
        from_polyline3d
    from ladybug_rhino.intersect import join_geometry_to_brep, join_geometry_to_mesh, \
        intersect_mesh_rays, bounding_box_extents, trace_ray, normal_at_point, \
        intersect_mesh_rays_distance
    from ladybug_rhino.grasshopper import all_required_inputs, list_to_data_tree, \
        hide_output, show_output
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component) and _run:
    # check the _bounce_count_
    _bounce_count_ = 0 if _bounce_count_ is None else _bounce_count_ - 1
    assert _bounce_count_ >= 0, 'The input _bounce_count_ must be greater '  \
        'than zero. Got {}.'.format(_bounce_count_ + 1)
    # process the input sun vector
    lb_vec = to_vector3d(_vector).normalize()
    neg_lb_vec = -lb_vec
    vec = from_vector3d(lb_vec)

    # convert all of the _source_geo and contex into a single Brep for ray tracing
    rtrace_brep = join_geometry_to_brep(_source_geo + context_)

    # autocompute the first and last bounce length if it's unspecified
    if _first_length_ is None or _last_length_ is None:
        max_pt, min_pt = (to_point3d(p) for p in bounding_box_extents(rtrace_brep))
        diag_dist = max_pt.distance_to_point(min_pt)
        _first_length_ = diag_dist if _first_length_ is None else _first_length_
        _last_length_ = diag_dist if _last_length_ is None else _last_length_

    # create the gridded mesh from the _source_geo and set up the starting rays
    study_mesh = to_joined_gridded_mesh3d(_source_geo, _grid_size)
    move_vec = neg_lb_vec * _first_length_
    source_points = [pt + move_vec for pt in study_mesh.face_centroids]
    source_normals = study_mesh.face_normals
    lb_rays = [Ray3D(pt, lb_vec) for pt in source_points]
    start_rays = [from_ray3d(ray) for ray in lb_rays]

    # if context_block_ is set to True, filter the source_points
    if context_block_:
        shade_mesh = join_geometry_to_mesh(context_)
        rev_vec = [from_vector3d(to_vector3d(_vector).reverse())]
        normals = [from_vector3d(vec) for vec in study_mesh.face_normals]
        points = [from_point3d(pt) for pt in study_mesh.face_centroids]
        int_matrix, angles = intersect_mesh_rays(
            shade_mesh, points, rev_vec, normals, cpu_count=1)
        new_start_rays, new_source_points, new_source_normals = [], [], []
        for ray, pt, norm, inter in zip(start_rays, source_points, source_normals, int_matrix):
            if inter[0] == 1:
                new_start_rays.append(ray)
                new_source_points.append(pt)
                new_source_normals.append(norm)
        start_rays, source_points, source_normals = \
            new_start_rays, new_source_points, new_source_normals

    # trace each ray through the geometry
    cutoff_ang = math.pi / 2
    rtrace_geo = [rtrace_brep]
    rays, int_pts = [], []
    for ray, pt, norm in zip(start_rays, source_points, source_normals):
        if norm.angle(neg_lb_vec) < cutoff_ang:
            pl_pts = trace_ray(ray, rtrace_geo, _bounce_count_ + 2)
            # if the intersection was successful, create a polyline represeting the ray
            if pl_pts:
                # gather all of the intersection points
                all_pts = [pt]
                for i_pt in pl_pts:
                    all_pts.append(to_point3d(i_pt))
                # compute the last point
                if len(pl_pts) < _bounce_count_ + 2:
                    int_norm = normal_at_point(rtrace_brep, pl_pts[-1])
                    int_norm = to_vector3d(int_norm)
                    last_vec = all_pts[-2] - all_pts[-1]
                    last_vec = last_vec.normalize()
                    final_vec = last_vec.reflect(int_norm).reverse()
                    final_pt = all_pts[-1] + (final_vec * _last_length_)
                    all_pts.append(final_pt)
                # create a Polyline3D from the points
                lb_ray_line = Polyline3D(all_pts)
                rays.append(from_polyline3d(lb_ray_line))
                int_pts.append([from_point3d(p) for p in all_pts])

    # if a receiver is specified, filter the output rays for intersection
    if len(receiver_) != 0:
        rec_mesh = join_geometry_to_mesh(receiver_)
        new_rays, new_int_pts = [], []
        for pl_pts, sun_ray in zip(int_pts, rays):
            start_point = pl_pts[-2]
            end_vec3d = to_point3d(pl_pts[-1]) - to_point3d(pl_pts[-2])
            end_vec = from_vector3d(end_vec3d.normalize())
            dist_list = intersect_mesh_rays_distance(rec_mesh, start_point, [end_vec])
            if 0 < dist_list[0] <= end_vec3d.magnitude + tolerance:
                new_rays.append(sun_ray)
                new_int_pts.append(pl_pts)
        rays, int_pts = new_rays, new_int_pts

    # convert the intersection points to a data tree
    int_pts = list_to_data_tree(int_pts)
    hide_output(ghenv.Component, 2)



# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2024, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Visualize the desirability of shade in terms of the time period of blocked
sun vectors for each part of a shade geometry.
_
The calculation assumes that all input _vectors represent sun to be blocked, which
is often the case when evaluating shade in terms of its benefit for glare reduction
and occupant visual comfort. It can also be the case when sun vectors have been
filtered to account for times of peak cooling demand or for the heat stress of
human subjects.
_
The component outputs a colored mesh of the shade illustrating the helpfulness of
shading each part of the _shade_geo. A higher saturation of blue indicates that
shading the cell blocks more hours of sun and is therefore more desirable.
_
The units for shade desirability are hrs/square Rhino unit, which note the amount
of time that sun is blocked by a given cell. So, if a given square meter of input
_shade_geo has a shade desirability of 10 hrs/m2, this means that a shade in
this location blocks an average of 10 hours to each of the _study_points.
-

    Args:
        _vectors: Sun vectors from the "LB SunPath" component, which will be used
            to determine the number of hours of sun blocked by the _shade_geo.
            When evaluating shade benefit in terms of glare reduction, these
            vectors are typically for any sun-up hour of the year since looking
            into the sun at practically any hour is likely to induce glare.
            When using this component to approximate reductions to cooling
            demand or human heat stress, it's more appropriate to filter
            sun vectors using a conditional statement or use other types of
            shade benefit analysis like the "LB Thermal Shade Benefit"
            component or the "HB Energy Shade Benefit" component.
        _study_points: Points representing an location in space for which shading
            desirability is being evaluated. For a study of shade desirability for
            reducing glare, this is often the location of the human subject's
            view. For a study of shade desirability over a surface like a desk
            or a window, the "LB Generate Point Grid" component can be used to
            create a set of points over the surface to input here.
        study_directs_: Optional Vectors that align with the _study_points and represent
            the direction in which shade desirability is being evaluated. For a
            study of shade desirability for reducing glare, this is the direction
            in which human subject is looking. For a study of shade desirability
            over a surface like a desk or a window, the vectors output of the
            "LB Generate Point Grid" component should be input here. If not
            supplied, sun vectors coming from any direction will be used to
            evualuate shade desirability.
        _shade_geo: Rhino Breps and/or Rhino Meshes representing shading to be evaluated
            in terms of its benefit. Note that, in the case that multiple
            shading geometries are connected, this component does not account
            for the interaction between the different shading surfaces and will
            just evaluate each part of the shade independently.
        context_: Rhino Breps and/or Rhino Meshes representing context geometry that can
            block sunlight to the _study_points, therefore discounting any benefit
            or harm that could come to the region.
        _grid_size: A positive number in Rhino model units for the size of grid cells at
            which the input _shade_geo will be subdivided for shade benefit
            analysis. The smaller the grid size, the higher the resolution of
            the analysis and the longer the calculation will take.  So it is
            recommended that one start with a large value here and decrease
            the value as needed. However, the grid size should usually be
            smaller than the dimensions of the smallest piece of the _shade_geo
            and context_ in order to yield meaningful results.
        _timestep_: A positive integer for the number of timesteps per hour at which
            the "LB SunPath" component generated sun vectors. This is used
            to correctly interpret the time duration represented by each of
            the input sun vectors. (Default: 1 for 1 vector per hour).
        legend_par_: Optional legend parameters from the "LB Legend Parameters"
            that will be used to customize the display of the results.
        _cpu_count_: An integer to set the number of CPUs used in the execution of the
            intersection calculation. If unspecified, it will automatically default
            to one less than the number of CPUs currently available on the
            machine or 1 if only one processor is available.
        _run: Set to "True" to run the component and perform shade benefit analysis.

    Returns:
        report: ...
        mesh: A colored mesh of the _shade_geo showing where shading is helpful (in blue),
            and where it does not make much of a difference (white or desaturated
            colors). Note that the colors can change depending upon the input legend_par_.
        legend: Legend showing the numeric values of hrs / square unit that correspond to
            the colors in the shade mesh.
        title: A text object for the study title.
        shade_help: The cumulative hrs / square unit helped by shading the given cell. If a given
            square meter of _shade_geo has a shade helpfulness of 10 hrs/m2,
            this means that a shade in this location blocks an average of 10 hours
            to each of the _study_points.
"""

ghenv.Component.Name = 'LB Shade Benefit'
ghenv.Component.NickName = 'ShadeBenefit'
ghenv.Component.Message = '1.8.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '3 :: Analyze Geometry'
ghenv.Component.AdditionalHelpFromDocStrings = '4'


try:
    from ladybug.color import Colorset
    from ladybug.graphic import GraphicContainer
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_rhino.config import units_abbreviation
    from ladybug_rhino.togeometry import to_joined_gridded_mesh3d, to_vector3d
    from ladybug_rhino.fromgeometry import from_mesh3d, from_point3d, from_vector3d
    from ladybug_rhino.fromobjects import legend_objects
    from ladybug_rhino.text import text_objects
    from ladybug_rhino.intersect import join_geometry_to_mesh, generate_intersection_rays, \
        intersect_rays_with_mesh_faces
    from ladybug_rhino.grasshopper import all_required_inputs, recommended_processor_count
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component) and _run:
    # set the defaults and process all of the inputs
    workers = _cpu_count_ if _cpu_count_ is not None else recommended_processor_count()
    _timestep_ = 1 if _timestep_ is None else _timestep_
    study_directs_ = None if len(study_directs_) == 0 else study_directs_

    # create the gridded mesh from the geometry
    analysis_mesh = to_joined_gridded_mesh3d(_shade_geo, _grid_size)
    mesh = from_mesh3d(analysis_mesh)

    # create a series of rays that represent the sun projected through the shade
    rev_vec = [from_vector3d(to_vector3d(vec).reverse()) for vec in _vectors]
    int_rays = generate_intersection_rays(_study_points, rev_vec)

    # if there is context, remove any rays that are blocked by the context
    shade_mesh = join_geometry_to_mesh(context_) \
        if len(context_) != 0 and context_[0] is not None else None

    # intersect the sun rays with the shade mesh
    face_int = intersect_rays_with_mesh_faces(
        mesh, int_rays, shade_mesh, study_directs_, cpu_count=workers)

    # loop through the face intersection result and evaluate the benefit
    pt_count = len(_study_points)
    shade_help = []
    for face_res, face_area in zip(face_int, analysis_mesh.face_areas):
        # Normalize by the area of the cell so there's is a consistent metric
        # between cells of different areas.
        # Also, divide the number of study points so people get a sense of
        # the average hours of blocked sun.
        shd_help = ((len(face_res) / face_area) / _timestep_) / pt_count
        shade_help.append(shd_help)

    # create the mesh and legend outputs
    graphic = GraphicContainer(shade_help, analysis_mesh.min, analysis_mesh.max, legend_par_)
    graphic.legend_parameters.title = 'hrs/{}2'.format(units_abbreviation())
    if legend_par_ is None or legend_par_.are_colors_default:
        graphic.legend_parameters.colors = Colorset.shade_benefit()
    if legend_par_ is None or legend_par_.min is None:
        graphic.legend_parameters.min = 0
    title = text_objects('Shade Benefit', graphic.lower_title_location,
                         graphic.legend_parameters.text_height * 1.5,
                         graphic.legend_parameters.font)

    # create all of the visual outputs
    analysis_mesh.colors = graphic.value_colors
    mesh = from_mesh3d(analysis_mesh)
    legend = legend_objects(graphic.legend)



# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2024, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Genrate a mesh with corresponding test points from a Rhino Brep (or Mesh).
_
The resulting mesh will be in a format that the "LB Spatial Heatmap" component
will accept.
-

    Args:
        _geometry: Brep or Mesh from which to generate the points and grid.
        _grid_size: Number for the size of the test grid.
        _offset_dist_: Number for the distance to move points from the surfaces
            of the input _geometry.  Typically, this should be a small positive
            number to ensure points are not blocked by the mesh. (Default: 0).
        quad_only_: Boolean to note whether meshing should be done using Rhino's
            defaults (False), which fills the entire _geometry to the edges
            with both quad and tringulated faces, or a mesh with only quad
            faces should be generated.
            _
            FOR ADVANCED USERS: This input can also be a vector object that will
            be used to set the orientation of the quad-only grid. Note that,
            if a vector is input here that is not aligned with the plane of
            the input _geometry, an error will be raised.

    Returns:
        points: Test points at the center of each mesh face.
        vectors: Vectors for the normal direction at each of the points.
        face_areas: Area of each mesh face.
        mesh: Analysis mesh that can be passed to the "LB Spatial Heatmap" component.
"""

ghenv.Component.Name = "LB Generate Point Grid"
ghenv.Component.NickName = 'GenPts'
ghenv.Component.Message = '1.8.1'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '4 :: Extra'
ghenv.Component.AdditionalHelpFromDocStrings = '1'

try:
    from ladybug_geometry.geometry3d.plane import Plane
    from ladybug_geometry.geometry3d.face import Face3D
    from ladybug_geometry.geometry3d.mesh import Mesh3D
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_rhino.togeometry import to_gridded_mesh3d, to_mesh3d, \
        to_face3d, to_vector3d
    from ladybug_rhino.fromgeometry import from_mesh3d, from_point3d, from_vector3d
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    # check the input and generate the mesh.
    _offset_dist_ = _offset_dist_ or 0
    if quad_only_:  # use Ladybug's built-in meshing methods
        lb_faces = to_face3d(_geometry)
        try:
            x_axis = to_vector3d(quad_only_)
            lb_faces = [Face3D(f.boundary, Plane(f.normal, f[0], x_axis), f.holes)
                        for f in lb_faces]
        except AttributeError:
            pass  # no plane connected; juse use default orientation
        lb_meshes = []
        for geo in lb_faces:
            try:
                lb_meshes.append(geo.mesh_grid(_grid_size, offset=_offset_dist_))
            except AssertionError:  # tiny geometry not compatible with quad faces
                continue
        if len(lb_meshes) == 0:
            lb_mesh = None
        elif len(lb_meshes) == 1:
            lb_mesh = lb_meshes[0]
        elif len(lb_meshes) > 1:
            lb_mesh = Mesh3D.join_meshes(lb_meshes)
    else:  # use Rhino's default meshing
        try:  # assume it's a Rhino Brep
            lb_mesh = to_gridded_mesh3d(_geometry, _grid_size, _offset_dist_)
        except Exception:  # assume it's a Rhino Mesh
            try:
                lb_mesh = to_mesh3d(_geometry)
            except Exception:  # unidientified geometry type
                raise TypeError(
                    '_geometry must be a Brep or a Mesh. Got {}.'.format(type(_geometry)))

    # generate the test points, vectors, and areas.
    if lb_mesh is not None:
        points = [from_point3d(pt) for pt in lb_mesh.face_centroids]
        vectors = [from_vector3d(vec) for vec in lb_mesh.face_normals]
        face_areas = lb_mesh.face_areas
        mesh = [from_mesh3d(lb_mesh)]
    else:
        mesh = []


# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2024, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Get the hours, days, or months of the year associated with the values of a data collection.
-

    Args:
        _data: An hourly, daily, or monthly collection from which hours, days, or
            months of the year will be retrieved.

    Returns:
        hoys: Numbers for the, hours, days or months of the year in the data collection.
"""

ghenv.Component.Name = 'LB Data DateTimes'
ghenv.Component.NickName = 'DataDT'
ghenv.Component.Message = '1.8.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '1 :: Analyze Data'
ghenv.Component.AdditionalHelpFromDocStrings = '0'

try:
    from ladybug.dt import DateTime
    from ladybug.datacollection import HourlyDiscontinuousCollection, DailyCollection, \
        MonthlyCollection, MonthlyPerHourCollection
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    if isinstance(_data, HourlyDiscontinuousCollection):
        hoys = [dt.hoy for dt in _data.datetimes]
    elif isinstance(_data, (MonthlyCollection, DailyCollection)):
        hoys = _data.datetimes
    elif isinstance(_data, MonthlyPerHourCollection):
        hoys = [DateTime(dt[0], 1, dt[1], dt[2]).hoy for dt in _data.datetimes]
    else:
        raise ValueError('Expected data collection. Got {}.'.format(type(data)))



# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2024, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Construct a Ladybug Header to be used to create a ladybug DataCollection.
-

    Args:
        _data_type: Text representing the type of data (e.g. Temperature). A full list
            of acceptable inputs can be seen by checking the all_u output of
            the "LB Unit Converter" component. This input can also be a custom
            DataType object that has been created with the "LB Construct Data
            Type" component.
        _unit_: Units of the data_type (e.g. C). Default is to use the
            base unit of the connected_data_type.
        _a_period: A Ladybug AnalysisPeriod object. (Default
        metadata_: Optional metadata to be associated with the Header. The input should
            be a list of text strings with a property name and value for the
            property separated by a colon. For example:
            _
            .    source: TMY
            .    city: New York
            .    country: USA

    Returns:
        header: A Ladybug Header object.
"""

ghenv.Component.Name = "LB Construct Header"
ghenv.Component.NickName = 'ConstrHeader'
ghenv.Component.Message = '1.8.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '1 :: Analyze Data'
ghenv.Component.AdditionalHelpFromDocStrings = '1'

try:
    import ladybug.datatype
    from ladybug.datatype.base import DataTypeBase
    from ladybug.header import Header
    from ladybug.analysisperiod import AnalysisPeriod
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))

# error message if the data type is not recognized
msg = 'The connected _data_type is not recognized.\nMake your own with ' \
    'the "LB Construct Data Type" component or choose from the following:' \
    '\n{}'.format('\n'.join(ladybug.datatype.BASETYPES))


if all_required_inputs(ghenv.Component):
    if isinstance(_data_type, DataTypeBase):
        pass
    elif isinstance(_data_type, str):
        _data_type = _data_type.replace(' ', '')
        try:
            _data_type = ladybug.datatype.TYPESDICT[_data_type]()
        except KeyError:  # check to see if it's a captilaization issue
            _data_type = _data_type.lower()
            for key in ladybug.datatype.TYPESDICT:
                if key.lower() == _data_type:
                    _data_type = ladybug.datatype.TYPESDICT[key]()
                    break
            else:
                raise TypeError(msg)
    else:
        raise TypeError(msg)

    if _unit_ is None:
        _unit_ = _data_type.units[0]

    if _a_period_ is None:
        _a_period_ = AnalysisPeriod()

    metadata_dict = {}
    if metadata_ != []:
        for prop in metadata_:
            key, value = prop.split(':')
            metadata_dict[key] = value.strip()

    header = Header(_data_type, _unit_, _a_period_, metadata_dict)


# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2024, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Draw an adaptive comfort chart in the Rhino scene and plot a set of prevailing and
indoor operative temperature values on it.
_
Connected data can include outdoor temperatures from imported EPW weather data
as well as indoor temperatures from an energy simulation.
-

    Args:
        _out_temp: Outdoor temperatures in one of the following formats:
            * A Data Collection of outdoor dry bulb temperatures recorded over
                the entire year. This Data Collection must be continouous and
                must either be an Hourly Collection or Daily Collection. In the event
                that the input adapt_par_ has a _avgm_or_runmean_ set to True,
                Monthly collections are also acceptable here. Note that, because
                an annual input is required, this input collection does not have
                to align with the _air_temp or _mrt_ inputs.
            * A Data Collection of prevailing outdoor temperature values in C.
                This Data Collection must align with the _air_temp or _mrt_
                inputs and bear the PrevailingOutdoorTemperature data type in
                its header.
            * A single prevailing outdoor temperature value in C to be used
                for all of the _air_temp or _mrt_ inputs.
        _air_temp: A hourly, daily, or sub-hourly data collection of temperature values
            for indoor air temperature (or operative temperature). If this input
            data collection is in in Farenheit, the entire chart will be drawn
            using IP units.
        _mrt_: A hourly, daily, or sub-hourly data collection for mean radiant temperature
            (MRT) in C. Default is the same as the air_temp, effectively
            interpreting the _air_temp as operative tempreature.
        _air_speed_: A number for the air speed values in m/s. Note that higher air speeds
            in the adaptive model only widen the upper boundary of the comfort
            range at temperatures above 24 C and will not affect the lower temperature
            of the comfort range. Default is a very low speed of 0.1 m/s.
        adapt_par_: Optional comfort parameters from the "LB Adaptive Comfort Parameters"
            component to specify the criteria under which conditions are
            considered acceptable/comfortable. The default will use ASHRAE-55
            adaptive comfort criteria.
        _base_pt_: A point to be used as the bottom-left-most point from which all
            geometry of the plot will be generated. (Default: (0, 0, 0)).
        _scale_: A number to set the dimensions of the chart. (Default: 1).
        _prevail_range_: An optional domain (or number for the upper temperature), which
            will be used to set the lower and upper boundaries of prevailing
            outdoor temperature on the chart. These should be in Celsius if the
            chart is plotted in Celsius and and Fahrenheit if it is in Fahrenheit.
            The default is 10C to 33C when ASHRAE-55 is used and 10C to 30C when
            the EN standard is used. This translates to 50F to 92F for ASHRAE-55
            and 50F to 86F for the EN standard.
        _operat_range_: An optional domain (or number for the upper temperature), which
            will be used to set the lower and upper boundaries of indoor
            operative temperature on the chart. These should be in Celsius if the
            chart is plotted in Celsius and and Fahrenheit if it is in Fahrenheit.
            The default is 14C to 40C, which translates to 58F to 104F.
        legend_par_: An optional LegendParameter object from the "LB Legend Parameters"
            component to change the display of the Adaptive Chart.
        data_: Optional data collections, which are aligned with the input _air_temp,
            which will be output from the data of this component and can be used
            to color points with data. This data can also be used along with
            the statement_ below to select out data that meets certain conditions.
        statement_: A conditional statement as a string (e.g. a > 25).
            .   
            The variable of the first data collection should always be named 'a'
            (without quotations), the variable of the second list should be
            named 'b', and so on.
            .
            For example, if three data collections are connected to _data
            and the following statement is applied:
            '10 < a < 30 and b < 33 and c > 2'
            The resulting collections will only include values where the first
            data collection is between 10 and 30, the second collection is less
            than 33 and the third collection is greater than 2.
            .
            For this component, the input indoor air temperature will always be
            the last (or seconf-to-last) letter and this will be followed by
            the input _mrt_ (if it is present).
        period_: A Ladybug analysis period to be applied to the _out_temp and
            _air_temp of the input data_.

    Returns:
        report: ...
        total_comfort: The percent of the data on the adaptive chart that is inside the
            comfort polygon.
        comfort_data: Data collection or a 0/1 value noting whether each of the data
            points on the chart lies inside of the comfort polygon.
            _
            This can be connected to the "LB Create Legend" component to generate
            a list of colors that can be used to color the points output from
            "LB Adaptive Chart" component to see exactly which points are
            comfortable and which are not.
            _
            Values are one of the following:
                0 = uncomfortable
                1 = comfortable
        condition_data: Data collection of integers noting the thermal status of the human
            subject according to the assigned comfort parameters.
            _
            This can be connected to the "LB Create Legend" component to generate
            a list of colors that can be used to color the points output from
            "LB Adaptive Chart" component to see exactly which points are
            hot, cold, and neutral.
            _
            Values are one of the following:
                * -1 = cold
                *  0 = netural
                * +1 = hot
        polygon: Brep representing the range of comfort for the input parameters.
        title: Text objects for the chart title and axes titles as well as a polyline for
            the outer border of the chart.
        prevail_lines: A list of line segments and text objects for the outdoor prevailing
            temperature labels on the chart.
        operative_lines: A list of line segments and text objects for the indoor operative
            temperature labels on the chart.
        mesh: A colored mesh showing the number of input hours that happen in each part
            of the adaptive chart.
        legend: A colored legend showing the number of hours that correspond to
            each color.
        points: Points representing each of the input prevailing and operative temperature values.
            By default, this ouput is hidden and it should be connected it to a
            native Grasshopper preview component to view it.
        data: The input data_ with the input statements or the periods applied to it.
            These can be deconstructed with the "LB Deconstruct Data" component
            and the resulting values can be plugged into the "LB Create Legend"
            component to generate colors that can be used to color the points
            above using the native Grasshopper "Custom Preview" component.
        vis_set: An object containing VisualizationSet arguments for drawing a detailed
            version of the Adaptive Chart in the Rhino scene. This can be
            connected to the "LB Preview Visualization Set" component to display
            this version of the Adaptive Chart in Rhino.
"""

ghenv.Component.Name = 'LB Adaptive Chart'
ghenv.Component.NickName = 'AdaptiveChart'
ghenv.Component.Message = '1.8.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '2 :: Visualize Data'
ghenv.Component.AdditionalHelpFromDocStrings = '2'

try:
    from ladybug_geometry.geometry2d import Point2D, Vector2D, LineSegment2D
    from ladybug_geometry.geometry3d import Point3D, Vector3D, Plane
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_geometry:\n\t{}'.format(e))

try:
    from ladybug.datacollection import BaseCollection
    from ladybug.datatype.temperature import PrevailingOutdoorTemperature
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_comfort.chart.adaptive import AdaptiveChart
    from ladybug_comfort.parameter.adaptive import AdaptiveParameter
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_comfort:\n\t{}'.format(e))

try:
    from ladybug_rhino.config import conversion_to_meters
    from ladybug_rhino.togeometry import to_point2d, to_point3d
    from ladybug_rhino.fromgeometry import from_mesh2d, from_polygon2d, \
        from_polyline2d, from_linesegment2d, from_point2d, \
        from_polyline2d_to_offset_brep
    from ladybug_rhino.text import text_objects
    from ladybug_rhino.fromobjects import legend_objects
    from ladybug_rhino.grasshopper import all_required_inputs, list_to_data_tree, \
        hide_output, longest_list, objectify_output
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


def leg_par_by_index(i):
    """Get legend parameters associated with a given index of the chart."""
    try:
        return legend_par_[i]
    except IndexError:
        return None


def small_labels(adapt_chart, labels, points, x_align, y_align, factor=1.0):
    """Translate a list of psych chart text labels into the Rhino scene."""
    return [text_objects(txt, Plane(o=Point3D(pt.x, pt.y, z)),
                         adapt_chart.legend_parameters.text_height * factor,
                         adapt_chart.legend_parameters.font, x_align, y_align)
            for txt, pt in zip(labels, points)]


def plane_from_point(point_2d, align_vec=Vector3D(1, 0, 0)):
    """Get a Plane from a Point2D.

    Args:
        point_2d: A Point2D to serve as the origin of the plane.
        align_vec: A Vector3D to serve as the X-Axis of the plane.
    """
    return Plane(o=Point3D(point_2d.x, point_2d.y, z), x=align_vec)


def draw_adapt_chart(adapt_chart):
    """Draw a given adaptive chart object into Rhino geometry.

    This will NOT translate any colored meshes or data points.
    """
    # output the comfort polygon and neutral lines
    polygon_i = [from_polyline2d_to_offset_brep([adapt_chart.comfort_polygon], offset, z)]
    nl = adapt_chart.neutral_polyline
    if isinstance(nl, LineSegment2D):
        polygon_i.append(from_linesegment2d(nl, z))
    else:
        polygon_i.append(from_polyline2d(nl, z))

    # output all of the lines/polylines for the various axes
    title_i = [from_polygon2d(adapt_chart.chart_border, z)]
    prev_lines_i = [from_linesegment2d(tpl, z) for tpl in adapt_chart.prevailing_lines]
    op_lines_i = [from_linesegment2d(tol, z) for tol in adapt_chart.operative_lines]

    # add the text to the various lines
    title_i.append(text_objects(
        adapt_chart.x_axis_text, plane_from_point(adapt_chart.x_axis_location),
        adapt_chart.legend_parameters.text_height * 1.5,
        adapt_chart.legend_parameters.font, 1, 0))
    title_i.append(text_objects(
        adapt_chart.y_axis_text, plane_from_point(adapt_chart.y_axis_location, Vector3D(0, 1)),
        adapt_chart.legend_parameters.text_height * 1.5,
        adapt_chart.legend_parameters.font, 1, 0))
    prev_lines_i = prev_lines_i + small_labels(
        adapt_chart, adapt_chart.prevailing_labels, adapt_chart.prevailing_label_points, 1, 0)
    op_lines_i = op_lines_i + small_labels(
        adapt_chart, adapt_chart.operative_labels, adapt_chart.operative_label_points, 0, 3)

    # add all of the objects to the base list
    polygon.append(polygon_i)
    title.append(title_i)
    prevail_lines.append(prev_lines_i)
    operative_lines.append(op_lines_i)


if all_required_inputs(ghenv.Component):
    # process the base point
    bp = to_point2d(_base_pt_) if _base_pt_ is not None else Point2D()
    z = to_point3d(_base_pt_).z if _base_pt_ is not None else 0

    # create lists to be filled with objects
    total_comfort = []
    comfort_data = []
    condition_data = []
    polygon = []
    title = []
    prevail_lines = []
    operative_lines = []
    mesh = []
    legend = []
    points = []
    data_colls = []
    vis_set = []

    # loop through the input temperatures and humidity and plot psych charts
    for j, temperature in enumerate(_air_temp):
        # process the out_temp and mrt inputs
        out_temp = longest_list(_out_temp, j)
        mrt = None
        if len(_mrt_) != 0:
            mrt = longest_list(_mrt_, j)

        # sense if the input temperature is in Farenheit
        use_ip = False
        if temperature.header.unit != 'C':  # convert to C and set chart to use_ip
            temperature = temperature.to_si()
            out_temp = out_temp.to_si()
            if mrt is not None:
                mrt = mrt.to_si()
            use_ip = True

        # set default values for the chart dimensions
        adapt_par = adapt_par_ if adapt_par_ is not None else AdaptiveParameter()
        _scale_ = 1.0 if _scale_ is None else _scale_
        xy_dim = _scale_ * 2 / conversion_to_meters()
        xy_dim = xy_dim * (9 / 5) if use_ip else xy_dim
        offset = xy_dim * 0.25
        if use_ip:
            tp_min, tp_max = (50, 92) if adapt_par.ashrae_or_en else (50, 86)
            to_min, to_max = 58, 104
        else:
            tp_min, tp_max = (10, 33) if adapt_par.ashrae_or_en else (10, 30)
            to_min, to_max = 14, 40
        if _prevail_range_ is not None:
            tp_min, tp_max = _prevail_range_
        if _operat_range_ is not None:
            to_min, to_max = _operat_range_
        y_move_dist = -xy_dim * 0.04 * j
        base_pt = bp.move(Vector2D(0, y_move_dist))

        # apply any analysis periods and conditional statements to the input collections
        original_temperature = temperature
        all_data = data_ + [temperature]
        if mrt is not None:
            all_data.append(mrt)
        if period_ is not None:
            all_data = [coll.filter_by_analysis_period(period_) for coll in all_data]
        if statement_ is not None:
            all_data = BaseCollection.filter_collections_by_statement(all_data, statement_)

        # create the adaptive chart object and draw it in the Rhino scene
        if mrt is not None:
            mrt = all_data.pop(-1)
        adapt_chart = AdaptiveChart.from_air_and_rad_temp(
            out_temp, all_data[-1], mrt, _air_speed_, adapt_par,
            leg_par_by_index(0), base_pt, xy_dim, xy_dim,
            tp_min, tp_max, to_min, to_max, use_ip=use_ip)
        draw_adapt_chart(adapt_chart)
        ttl_tp = adapt_chart.container.lower_title_location.move(
            Vector3D(0, -adapt_chart.legend_parameters.text_height * 3))
        if z != 0:
            ttl_tp = Plane(n=ttl_tp.n, o=Point3D(ttl_tp.o.x, ttl_tp.o.y, z), x=ttl_tp.x)
        title[j + j * len(data_)].append(text_objects(
            adapt_chart.title_text, ttl_tp,
            adapt_chart.legend_parameters.text_height * 1.5,
            adapt_chart.legend_parameters.font, 0, 0))
        vs_args = [adapt_chart]

        # plot the data on the chart
        lb_points = adapt_chart.data_points
        points.append([from_point2d(pt) for pt in lb_points])
        hide_output(ghenv.Component, 10)
        mesh.append(from_mesh2d(adapt_chart.colored_mesh, z))
        leg = adapt_chart.legend
        if z != 0 and leg.legend_parameters.is_base_plane_default:
            nl_par = leg.legend_parameters.duplicate()
            m_vec = Vector3D(0, 0, z)
            nl_par.base_plane = nl_par.base_plane.move(m_vec)
            leg._legend_par = nl_par
        legend.append(legend_objects(leg))

        # process the comfort-related outputs
        total_comfort.append(adapt_chart.percent_comfortable)
        comfort_data.append(adapt_chart.is_comfortable)
        condition_data.append(adapt_chart.thermal_condition)

        # process any of the connected data into a legend and colors
        if len(data_) != 0:
            data_colls.append(all_data[:-1])
            move_dist = xy_dim * (adapt_chart.max_prevailing - adapt_chart.min_prevailing + 20)
            vs_leg_par = []
            for i, d in enumerate(all_data[:-1]):
                # create a new adaptive chart offset from the original
                new_pt = Point2D(base_pt.x + move_dist * (i + 1), base_pt.y)
                adapt_chart = AdaptiveChart.from_air_and_rad_temp(
                    out_temp, all_data[-1], mrt, _air_speed_, adapt_par,
                    leg_par_by_index(0), new_pt, xy_dim, xy_dim,
                    tp_min, tp_max, to_min, to_max, use_ip=use_ip)
                draw_adapt_chart(adapt_chart)
                lb_mesh, container = adapt_chart.data_mesh(d, leg_par_by_index(i + 1))
                mesh.append(from_mesh2d(lb_mesh, z))
                leg = container.legend
                vs_leg_par.append(leg.legend_parameters)
                if z != 0 and leg.legend_parameters.is_base_plane_default:
                    nl_par = leg.legend_parameters.duplicate()
                    m_vec = Vector3D(0, 0, z)
                    nl_par.base_plane = nl_par.base_plane.move(m_vec)
                    leg._legend_par = nl_par
                legend.append(legend_objects(leg))
                move_vec = Vector2D(base_pt.x + move_dist * (i + 1), 0)
                points.append([from_point2d(pt.move(move_vec)) for pt in lb_points])

                # add a title for the new chart
                title_items = ['Adaptive Chart'] +\
                    ['{} [{}]'.format(d.header.data_type, d.header.unit)] + \
                    ['{}: {}'.format(key, val) for key, val in d.header.metadata.items()]
                ttl_tp = adapt_chart.container.lower_title_location.move(
                    Vector3D(0, -adapt_chart.legend_parameters.text_height * 3))
                if z != 0:
                    ttl_tp = Plane(n=ttl_tp.n, o=Point3D(ttl_tp.o.x, ttl_tp.o.y, z), x=ttl_tp.x)
                title[j + j * len(data_) + (i + 1)].append(text_objects(
                    '\n'.join(title_items), ttl_tp,
                    adapt_chart.legend_parameters.text_height * 1.5,
                    adapt_chart.legend_parameters.font, 0, 0))
            vs_args.extend([all_data[:-2], vs_leg_par, z])
        else:
            vs_args.extend([None, None, z])
        vis_set.append(vs_args)

    # upack all of the python matrices into data trees
    polygon = list_to_data_tree(polygon)
    title = list_to_data_tree(title)
    prevail_lines = list_to_data_tree(prevail_lines)
    operative_lines = list_to_data_tree(operative_lines)
    legend = list_to_data_tree(legend)
    points = list_to_data_tree(points)
    data = list_to_data_tree(data_colls)

    # output arguments for the visualization set
    vis_set = objectify_output('VisualizationSet Aruments [AdaptiveChart]', vis_set)



# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2024, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>


"""
Create a custom legend for any set of data or range. Creating a legend with this
component allows for a bit more flexibility than what can be achieved by working
with the legends automatically output from different studies.
-

    Args:
        _values: A list of numerical values or data collections that the legend refers
            to. This can also be the minimum and maximum numerical values of the
            data. The legend's maximum and minimum values will be set by the max
            and min of the data set.
        _base_plane_: An optional plane or point to set the location of the
            legend. (Default: Rhino origin - (0, 0, 0))
        title_: A text string representing a legend title. Legends are usually
            titled with the units of the data.
        legend_par_: Optional legend parameters from the "LB Legend Parameters" component.
        leg_par2d_: Optional 2D LegendParameters from the "LB Legend Parameters 2D"
            component, which will be used to customize a legend in the plane
            of the screen so that it functions like a head-up display (HUD).
            If unspecified, the VisualizationSet will be rendered with 3D
            legends in the Rhino scene much like the other native Ladybug
            Tools components.

    Returns:
        mesh: A colored mesh for the legend colors.
        title_obj: A text object for the  legend title.
        label_objs: An array of text objects for the label text.
        label_text: An array of text strings for the label text.
        colors: An array of colors that align with the input _values. This can
            be used to color geometry that aligns with the values.
"""

from ghpythonlib.componentbase import executingcomponent as component
import Grasshopper, GhPython
import System
import Rhino
import rhinoscriptsyntax as rs

class MyComponent(component):
    
    def __init__(self):
        super(MyComponent,self).__init__()
        self.draw_2d_text = None
        self.draw_sprite = None
        self.colored_mesh = None
    
    def RunScript(self, _values, _base_plane_, title_, legend_par_, leg_par2d_):
        ghenv.Component.Name = "LB Create Legend"
        ghenv.Component.NickName = 'CreateLegend'
        ghenv.Component.Message = '1.8.0'
        ghenv.Component.Category = 'Ladybug'
        ghenv.Component.SubCategory = '4 :: Extra'
        ghenv.Component.AdditionalHelpFromDocStrings = '0'
        
        try:
            from ladybug.legend import Legend, LegendParameters
        except ImportError as e:
            raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))
        
        try:
            from ladybug_rhino.togeometry import to_plane
            from ladybug_rhino.fromobjects import legend_objects
            from ladybug_rhino.color import color_to_color
            from ladybug_rhino.preview import VisualizationSetConverter
            from ladybug_rhino.grasshopper import all_required_inputs
        except ImportError as e:
            raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))
        
        
        if all_required_inputs(ghenv.Component):
            # set default values
            legend_par_ = legend_par_.duplicate() if legend_par_ is not None else \
                LegendParameters()
            if _base_plane_:
                legend_par_.base_plane = to_plane(_base_plane_)
            legend_par_.title = title_
            
            # create the legend
            values = []
            for val in _values:
                try:
                    values.append(float(val))
                except AttributeError:  # assume it's a data collection
                    values.extend(val.values)
            legend = Legend(values, legend_par_)
            colors = [color_to_color(col) for col in legend.value_colors]
            label_text = legend.segment_text
            
            if leg_par2d_ is None:  # output a 3D legend
                self.draw_2d_text, self.draw_sprite = None, None
                rhino_objs = legend_objects(legend)
                mesh = rhino_objs[0]
                title_obj = rhino_objs[1]
                label_objs = rhino_objs[2:]
            else:  # output a 2D legend that is oriented to the screen
                mesh, title_obj, label_objs= None, None, None
                legend.legend_parameters.properties_2d = leg_par2d_
                d_sprite, self.draw_2d_text = \
                    VisualizationSetConverter.convert_legend2d(legend)
                self.draw_sprite = [d_sprite]
        else:
            mesh, title_obj, label_objs, label_text, colors = \
                None, None, None, None, None
            self.draw_2d_text, self.draw_sprite = None, None
        
        # return outputs if you have them; here I try it for you
        self.colored_mesh = mesh
        return (mesh, title_obj, label_objs, label_text, colors)
        
    def DrawViewportMeshes(self, args):
        try:
            # get the DisplayPipeline from the event arguments
            display = args.Display
            # draw the objects in the scene
            if self.colored_mesh is not None:
                display.DrawMeshFalseColors(self.colored_mesh)
            if self.draw_2d_text is not None:
                for draw_args in self.draw_2d_text:
                    display.Draw2dText(*draw_args)
            if self.draw_sprite is not None:
                for draw_args in self.draw_sprite:
                    display.DrawSprite(*draw_args)
        except Exception, e:
            System.Windows.Forms.MessageBox.Show(str(e), "script error")
    
    def get_ClippingBox(self):
        return Rhino.Geometry.BoundingBox()



# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2024, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Use this component to access a library of typical gradients useful throughout Ladybug. 
The output from this component should be plugged into the colors_ input of the
"Legend Parameters" component.
-
Note that the colorblind friendly schemes have prioritized readability for red-green
colorblindness (deuteranomaly, protanomaly, protanopia, and deuteranopia), which
is by far more common than blue-yellow colorblindness. However, they are not
necessarily ideal for all types of color blindness, though they are monotonic
and perceptually uniform to all forms of color vision. This means that they should
be readable as a dark-to-light scale by anyone.
-
For an image of each of the gardients in the library, check here:
https://github.com/ladybug-tools/lbt-grasshopper/blob/master/gradients.png
-

    Args:
        _index_: An index refering to one of the following possible gradients:
            0 - Original Ladybug
            1 - Nuanced Ladybug
            2 - Multi-colored Ladybug
            3 - Ecotect
            4 - View Study
            5 - Shadow Study
            6 - Glare Study
            7 - Annual Comfort
            8 - Thermal Comfort
            9 - Peak Load Balance
            10 - Heat Sensation
            11 - Cold Sensation
            12 - Benefit/Harm
            13 - Harm
            14 - Benefit
            15 - Shade Benefit/Harm
            16 - Shade Harm
            17 - Shade Benefit
            18 - Energy Balance
            19 - Energy Balance w/ Storage
            20 - THERM
            21 - Cloud Cover
            22 - Black to White
            23 - Blue, Green, Red
            24 - Multicolored 2
            25 - Multicolored 3
            26 - OpenStudio Palette
            27 - Cividis (colorblind friendly)
            28 - Viridis (colorblind friendly)
            29 - Parula (colorblind friendly)

    Returns:
        colors: A series of colors to be plugged into the "LB Legend Parameters"
            component.
"""

ghenv.Component.Name = 'LB Color Range'
ghenv.Component.NickName = 'ColRange'
ghenv.Component.Message = '1.8.1'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '4 :: Extra'
ghenv.Component.AdditionalHelpFromDocStrings = '1'

try:
    from ladybug.color import Colorset
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_rhino.color import color_to_color
    from ladybug_rhino.grasshopper import turn_off_old_tag
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))
turn_off_old_tag(ghenv.Component)


_index_ = _index_ or 0
cs = Colorset()
colors = [color_to_color(col) for col in cs[_index_]]



# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2024, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Produce a DDY file from the data contained within an EPW or STAT file.
_
For EPW files, this method will first check if there is any heating or cooling
design day information contained within the EPW itself. If None is found, the
heating and cooling design days will be derived from analysis of the annual
data within the EPW. This process of analyzing the annual TMY data is
less representative of the climate since only one year of data is used to
derive the DDY (instead of the usual multi-year analysis). However, if the
EPW is the best available representation of the climate for a given site, it
can often be preferable to using a DDY constructed with more years of data
but from further away. Information on the uncertainty introduced by using
only one year of data to create design days can be found in AHSRAE HOF 2013,
Chapter 14.14.
_
For STAT files, the DDY file will only be produced if the design day information
is found within the file. If no information on the relevant design days are
found, and error will be raised and the component will fail to run.
-

    Args:
        _weather_file: The path to an .epw file or .stat file on your system, from
            which a .ddy will be generated.
        _percentile_: A number between 0 and 50 for the percentile difference
            from the most extreme conditions within the EPW or STAT
            to be used for the design day. Typical values are 0.4
            and 1.0. (Default: 0.4).
        monthly_cool_: A boolean to note whether the resulting .ddy file should
            contain twelve cooling design days for each of the months of
            the year. This type of DDY file is useful when the peak cooling
            might not be driven by warm outdoor temperatures but instead by
            the highest-intensity solar condition, which may not conincide
            with the highest temperature. Monthly conditions will be for the
            2.0% hottest conditions in each month, which generally aligns with
            the annual 0.4% cooling design conditions.
        _folder_: An optional file path to a directory into which the DDY file
            will be written.  If None, the DDY file will be written to the
            ladybug default weather data folder and placed in a sub-folder
            called "ddy".
        _write: Set to "True" to write the .ddy file.

    Returns:
        ddy_file: A .ddy file path that has been written to your system.
"""

ghenv.Component.Name = 'LB EPW to DDY'
ghenv.Component.NickName = 'EPWtoDDY'
ghenv.Component.Message = '1.8.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '0 :: Import'
ghenv.Component.AdditionalHelpFromDocStrings = '0'

import os

try:
    from ladybug.epw import EPW
    from ladybug.stat import STAT
    from ladybug.config import folders
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component) and _write:
    # set default values
    _percentile_ = 0.4 if _percentile_ is None else _percentile_
    _folder_ = os.path.join(folders.default_epw_folder, 'ddy') if _folder_ \
        is None else _folder_
    if _weather_file.lower().endswith('.epw'):
        f_name = os.path.basename(_weather_file.lower()).replace('.epw', '.ddy')
        epw_data = True
    elif _weather_file.lower().endswith('.stat'):
        f_name = os.path.basename(_weather_file.lower()).replace('.stat', '.ddy')
        epw_data = False
    else:
        raise ValueError('Failed to recognize the file type of "{}".\n'
                         'Must end in .epw or .stat.'.format(_weather_file))
    f_path = os.path.join(_folder_, f_name)

    # create the DDY file
    if epw_data:
        epw = EPW(_weather_file)
        ddy_file = epw.to_ddy_monthly_cooling(f_path, _percentile_, 2) \
            if monthly_cool_ else epw.to_ddy(f_path, _percentile_)
    else:
        stat = STAT(_weather_file)
        ddy_file = stat.to_ddy_monthly_cooling(f_path, _percentile_, 2) \
            if monthly_cool_ else stat.to_ddy(f_path, _percentile_)



# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2024, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Orient a series of geometries to the active viewport camera.
-

    Args:
        _geo: A series of geometries to be oriented to the camera of the active
            Rhino viewport.
        _position_: A point to be used as the origin around which the the geometry
            will be oriented. If None, the lower left corner of the bounding box
            around the geometry will be used.
        refresh_: Connect a Grasshopper "button" component to refresh the orientation
            upon hitting the button. You can also connect a Grasshopper "Timer"
            component to update the view in real time as you navigate through
            the Rhino scene.
    
    Returns:
        geo: The input geometry that has been oriented to the camera of the active
            Rhino viewport.
"""

ghenv.Component.Name = 'LB Orient to Camera'
ghenv.Component.NickName = 'OrientCam'
ghenv.Component.Message = '1.8.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '4 :: Extra'
ghenv.Component.AdditionalHelpFromDocStrings = '2'

try:
    from ladybug_rhino.grasshopper import all_required_inputs
    from ladybug_rhino.viewport import orient_to_camera
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    geo = orient_to_camera(_geo, _position_)
